# pisek cms - Tool for importing tasks from Pisek into CMS.
#
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# The CMS module imports and calls internal CMS libraries,
# as well as dependencies of CMS (sqlalchemy).
# Since these dependencies aren't declared anywhere, importing this module
# will cause an exception on most installations.
# It is therefore important not to import this module unless a cms command has actually been called.

# The implementation uses CMS's sqlalchemy types to make changes to the database.
# It also uses some other APIs, like session creation.
# If CMS changes its database schema or some of its internal APIs,
# this module will likely break.

# This module obviously also uses internal methods of pisek,
# mostly config parsing and some general utilities.
# Large refactors in pisek will also require changes to this module.

# Some functionality is duplicated, most notably enumerating all inputs
# and finding the file extensions of solutions.
# All of this functionality is implemented inside private methods of pipeline jobs,
# making it inaccessible to part of Pisek that don't use the pipeline.

# The pipeline is called when creating a dataset, in order to generate inputs and outputs
# and to compile the judge.
# It is called as if the user requested the primary solution to be tested.

# One small thing to note is that which automatic submission corresponds
# to which solution isn't explicitly stored anywhere.
# Instead, the check and testing-log commands simply search the database for
# a matching (task, file hash, language) triple.

from argparse import Namespace
from cms.db.session import Session
from sqlalchemy.exc import IntegrityError

from pisek.cms.dataset import create_dataset, get_dataset
from pisek.cms.result import create_testing_log, check_results
from pisek.cms.submission import get_participation, submit_all
from pisek.cms.task import create_task, get_task, set_task_settings
from pisek.env.env import Env
from pisek.jobs.cache import Cache
from pisek.jobs.task_pipeline import TaskPipeline
from pisek.utils.pipeline_tools import with_env


def prepare_files(env: Env):
    contest_type = env.config.contest_type

    if contest_type != "cms":
        raise RuntimeError(f"Cannot upload {contest_type}-type task to CMS")

    env = env.fork()
    env.solutions = [env.config.primary_solution]

    if TaskPipeline(env).run_jobs(Cache(env), env) != 0:
        raise RuntimeError("Failed to test primary solution, cannot upload to CMS")


@with_env
def create(env: Env, args: Namespace) -> int:
    prepare_files(env)

    session = Session()

    description = args.description

    task = create_task(session, env, description)
    dataset = task.active_dataset

    try:
        session.commit()
    except IntegrityError as e:
        raise RuntimeError(
            "Failed to commit transaction, does the task already exist?"
        ) from e

    print(
        f'Created task {task.name} (id {task.id}) with dataset "{dataset.description}" (id {dataset.id})'
    )
    return 0


@with_env
def update(env: Env, args: Namespace) -> int:
    session = Session()

    task = get_task(session, env.config)
    set_task_settings(task, env.config)

    session.commit()

    print(f"Updated task {task.name} (id {task.id})")
    return 0


@with_env
def add(env: Env, args: Namespace) -> int:
    prepare_files(env)

    session = Session()

    description = args.description
    autojudge = not args.no_autojudge

    task = get_task(session, env.config)
    dataset = create_dataset(session, env, task, description, autojudge)

    try:
        session.commit()
    except IntegrityError as e:
        raise RuntimeError(
            "Failed to commit transaction, does a dataset with this description exist already?"
        ) from e

    print(f'Added dataset "{dataset.description}" (id {dataset.id})')
    return 0


@with_env
def submit(env: Env, args: Namespace) -> int:
    session = Session()

    username = args.username

    task = get_task(session, env.config)
    participation = get_participation(session, task, username)
    submissions = submit_all(session, env, task, participation)

    session.commit()

    for solution, submission in submissions:
        print(f"Submitted {solution} with id {submission.id}")
    return 0


@with_env
def testing_log(env: Env, args: Namespace) -> int:
    session = Session()

    description = args.dataset

    task = get_task(session, env.config)
    dataset = get_dataset(session, task, description)
    success = create_testing_log(session, env, dataset)

    return 0 if success else 1


@with_env
def check(env: Env, args: Namespace) -> int:
    session = Session()

    description = args.dataset

    task = get_task(session, env.config)
    dataset = get_dataset(session, task, description)
    success = check_results(session, env, dataset)

    return 0 if success else 1



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
import os
from typing import Optional, TYPE_CHECKING
import yaml
from pisek.utils.terminal import colored_env

if TYPE_CHECKING:
    from pisek.env.env import Env

BUILD_DIR = "build/"

GENERATED_SUBDIR = "generated/"
INPUTS_SUBDIR = "inputs/"
INVALID_OUTPUTS_SUBDIR = "invalid/"
OUTPUTS_SUBDIR = "outputs/"
SANITIZED_SUBDIR = "sanitized/"
LOG_SUBDIR = "log/"


@dataclass(frozen=True)
class TaskPath:
    """Class representing a path to task file."""

    path: str
    name: str

    def __init__(self, *path: str):
        joined_path = os.path.normpath(os.path.join(*path))
        object.__setattr__(self, "path", joined_path)
        object.__setattr__(self, "name", os.path.basename(joined_path))

    def __format__(self, __format_spec: str) -> str:
        match __format_spec:
            case "":
                return self.path
            case "p":
                return self.path
            case "n":
                return self.name
            case _:
                raise ValueError(
                    f"Invalid format specifier '{__format_spec}' for object of type '{self.__class__.__name__}'"
                )

    def __eq__(self, other_path) -> bool:
        if isinstance(other_path, TaskPath):
            return self.path == other_path.path
        else:
            return False

    def col(self, env: "Env") -> str:
        return colored_env(self.path, "magenta", env)

    def replace_suffix(self, new_suffix: str) -> "TaskPath":
        path = os.path.splitext(self.path)[0] + new_suffix
        return TaskPath(path)

    @staticmethod
    def from_abspath(*path: str) -> "TaskPath":
        return TaskPath(os.path.relpath(os.path.join(*path), "."))

    @staticmethod
    def static_path(env: "Env", *path: str) -> "TaskPath":
        return TaskPath(env.config.static_subdir, *path)

    @staticmethod
    def solution_path(env: "Env", *path: str) -> "TaskPath":
        return TaskPath(env.config.solutions_subdir, *path)

    @staticmethod
    def executable_path(env: "Env", *path: str) -> "TaskPath":
        return TaskPath(BUILD_DIR, *path)

    @staticmethod
    def executable_file(env: "Env", program: str) -> "TaskPath":
        program = os.path.splitext(os.path.basename(program))[0]
        return TaskPath.executable_path(env, program)

    @staticmethod
    def data_path(env: "Env", *path: str) -> "TaskPath":
        return TaskPath(env.config.data_subdir, *path)

    @staticmethod
    def generated_path(env: "Env", *path: str) -> "TaskPath":
        return TaskPath.data_path(env, GENERATED_SUBDIR, *path)

    @staticmethod
    def generated_input_file(env: "Env", subtask: int, seed: int) -> "TaskPath":
        return TaskPath.generated_path(env, f"{subtask:02}_{seed:x}.in")

    @staticmethod
    def input_path(env: "Env", *path: str) -> "TaskPath":
        return TaskPath.data_path(env, INPUTS_SUBDIR, *path)

    @staticmethod
    def invalid_path(env: "Env", *path: str) -> "TaskPath":
        return TaskPath.data_path(env, INVALID_OUTPUTS_SUBDIR, *path)

    @staticmethod
    def invalid_file(env: "Env", name: str, seed: int) -> "TaskPath":
        name = os.path.splitext(name)[0]
        return TaskPath.invalid_path(env, f"{name}.{seed:x}.invalid")

    @staticmethod
    def output_path(env: "Env", *path: str) -> "TaskPath":
        return TaskPath.data_path(env, OUTPUTS_SUBDIR, *path)

    @staticmethod
    def output_file(env: "Env", input_name: str, solution: str) -> "TaskPath":
        input_name = os.path.splitext(os.path.basename(input_name))[0]
        solution = os.path.basename(solution)
        return TaskPath.output_path(env, f"{input_name}.{solution}.out")

    @staticmethod
    def output_static_file(env: "Env", name: str) -> "TaskPath":
        name = os.path.splitext(name)[0]
        return TaskPath.output_path(env, f"{name}.out")

    @staticmethod
    def sanitized_path(env: "Env", *path: str) -> "TaskPath":
        return TaskPath.data_path(env, SANITIZED_SUBDIR, *path)

    @staticmethod
    def sanitized_file(env: "Env", name: str) -> "TaskPath":
        name = os.path.basename(name)
        return TaskPath.sanitized_path(env, f"{name}.clean")

    @staticmethod
    def log_path(env: "Env", *path: str) -> "TaskPath":
        return TaskPath.data_path(env, LOG_SUBDIR, *path)

    @staticmethod
    def points_file(env: "Env", name: str) -> "TaskPath":
        name = os.path.splitext(name)[0]
        return TaskPath.log_path(env, f"{name}.points")

    @staticmethod
    def log_file(env: "Env", name: Optional[str], program: str) -> "TaskPath":
        program = os.path.basename(program)
        if name is None:
            return TaskPath.log_path(env, f"{program}.log")

        name = os.path.splitext(os.path.basename(name))[0]
        return TaskPath.log_path(env, f"{name}.{program}.log")


def task_path_representer(dumper, task_path: TaskPath):
    return dumper.represent_sequence("!TaskPath", [task_path.path])


def task_path_constructor(loader, value):
    [path] = loader.construct_sequence(value)
    return TaskPath(path)


yaml.add_representer(TaskPath, task_path_representer)
yaml.add_constructor("!TaskPath", task_path_constructor)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import abstractmethod
from functools import cache
import os
import random
import subprocess
from typing import Optional, Union, Callable

from pisek.env.env import Env
from pisek.utils.paths import TaskPath
from pisek.config.config_types import ProgramType, OutCheck, JudgeType
from pisek.jobs.jobs import State, Job, PipelineItemFailure
from pisek.utils.text import tab
from pisek.jobs.parts.task_job import TaskJobManager
from pisek.jobs.parts.run_result import RunResult, RunResultKind
from pisek.jobs.parts.program import ProgramsJob
from pisek.jobs.parts.compile import Compile
from pisek.jobs.parts.chaos_monkey import Incomplete, ChaosMonkey
from pisek.jobs.parts.tools import Sanitize
from pisek.jobs.parts.solution_result import Verdict, SolutionResult

DIFF_NAME = "diff.sh"


class JudgeManager(TaskJobManager):
    """Manager that prepares and test judge."""

    def __init__(self) -> None:
        super().__init__("Preparing judge")

    def _get_jobs(self) -> list[Job]:
        jobs: list[Job] = []
        if self._env.config.out_check == OutCheck.judge:
            if self._env.config.out_judge is None:
                raise RuntimeError(
                    f"Unset judge for out_check={self._env.config.out_check.name}"
                )
            jobs.append(
                comp := Compile(self._env, TaskPath(self._env.config.out_judge))
            )

        samples = self._get_samples()
        if self._env.config.task_type == "communication":
            return jobs

        for inp, out in samples:
            jobs.append(
                judge_j := judge_job(
                    inp,
                    out,
                    out,
                    0,
                    lambda: "0",
                    1.0,
                    self._env,
                )
            )
            if self._env.config.out_check == OutCheck.judge:
                judge_j.add_prerequisite(comp)

            if os.stat(out.path).st_size > 0:
                JOBS = [(Incomplete, 10), (ChaosMonkey, 50)]

                total = sum(map(lambda x: x[1], JOBS))
                random.seed(4)  # Reproducibility!
                seeds = random.sample(range(0, 16**4), total)

                for job, times in JOBS:
                    for _ in range(times):
                        seed = seeds.pop()
                        inv_out = TaskPath.invalid_file(self._env, out.name, seed)
                        jobs += [
                            invalidate := job(self._env, out, inv_out, seed),
                            run_judge := judge_job(
                                inp,
                                inv_out,
                                out,
                                0,
                                lambda: "0",
                                None,
                                self._env,
                            ),
                        ]
                        if self._env.config.out_check == OutCheck.judge:
                            run_judge.add_prerequisite(comp)
                        run_judge.add_prerequisite(invalidate)
        return jobs


class RunKasiopeaJudgeMan(TaskJobManager):
    def __init__(
        self,
        subtask: int,
        seed: int,
        input_: str,
        output: str,
        correct_output: str,
    ) -> None:
        self._subtask = subtask
        self._seed = seed
        self._input = input_
        self._output = output
        self._correct_output = correct_output
        super().__init__("Running judge")

    def _get_jobs(self) -> list[Job]:
        input_, output, correct_output = map(
            TaskPath,
            (self._input, self._output, self._correct_output),
        )
        clean_output = TaskPath.sanitized_file(self._env, output.name)

        jobs: list[Job] = [
            sanitize := Sanitize(self._env, output, clean_output),
            judge := judge_job(
                input_,
                clean_output,
                correct_output,
                self._subtask,
                lambda: f"{self._seed:x}",
                None,
                self._env,
            ),
        ]
        judge.add_prerequisite(sanitize)
        if self._env.config.out_check == OutCheck.judge:
            if self._env.config.out_judge is None:
                raise RuntimeError(
                    f"Unset judge for out_check={self._env.config.out_check.name}"
                )
            jobs.insert(
                0,
                compile_judge := Compile(
                    self._env, TaskPath(self._env.config.out_judge)
                ),
            )
            judge.add_prerequisite(compile_judge)

        self._judge_job = judge

        return jobs

    def judge_result(self) -> SolutionResult:
        if self.state != State.succeeded:
            raise RuntimeError("Judging hasn't successfully finished.")
        elif not isinstance(self._judge_job.result, SolutionResult):
            raise RuntimeError("Judging result invalid.")
        return self._judge_job.result


JUDGE_JOB_NAME = r"Judge (\w+)"


class RunJudge(ProgramsJob):
    """Runs judge on single input. (Abstract class)"""

    def __init__(
        self,
        env: Env,
        name: str,
        judge_name: str,
        input_: TaskPath,
        judge_log_file: TaskPath,
        expected_points: Optional[float],
        **kwargs,
    ) -> None:
        super().__init__(env=env, name=name, **kwargs)
        self.input = input_
        self.judge_name = judge_name
        self.judge_log_file = judge_log_file
        self.expected_points = expected_points

        self.result: Optional[SolutionResult]

    @cache
    def _load_solution_run_res(self) -> None:
        """
        Loads solution's RunResult into self.solution_res
        """
        self._solution_run_res = self._get_solution_run_res()

    @abstractmethod
    def _get_solution_run_res(self) -> RunResult:
        """
        Gets solution's RunResult.
        Call this only through _load_solution_run_res as this can run the solution.
        """
        pass

    @abstractmethod
    def _judge(self) -> SolutionResult:
        """Here actually do the judging."""
        pass

    @abstractmethod
    def _judging_message(self) -> str:
        pass

    def _judging_message_capitalized(self) -> str:
        msg = self._judging_message()
        return msg[0].upper() + msg[1:]

    def _run(self) -> SolutionResult:
        self._load_solution_run_res()
        if self._solution_run_res.kind == RunResultKind.OK:
            result = self._judge()
        elif self._solution_run_res.kind == RunResultKind.RUNTIME_ERROR:
            result = SolutionResult(Verdict.error, 0.0, self._solution_run_res, None)
        elif self._solution_run_res.kind == RunResultKind.TIMEOUT:
            result = SolutionResult(Verdict.timeout, 0.0, self._solution_run_res, None)

        if (
            self.expected_points is not None
            and result is not None
            and result.points != self.expected_points
        ):
            raise PipelineItemFailure(
                f"{self._judging_message_capitalized()} should have got {self.expected_points} points but got {result.points} points."
            )

        return result

    def message(self) -> str:
        """Message about how judging ended."""
        if self.result is None:
            raise RuntimeError(f"Job {self.name} has not finished yet.")

        sol_rr = self.result.solution_rr
        judge_rr = self.result.judge_rr

        text = f"input: {self._quote_file_with_name(self.input)}"
        if isinstance(self, RunBatchJudge):
            text += f"correct output: {self._quote_file_with_name(self.correct_output)}"
        text += f"result: {self.result.verdict.name}\n"

        text += "solution:\n"
        text += tab(
            self._format_run_result(
                sol_rr,
                status=sol_rr.kind != RunResultKind.OK,
                stderr_force_content=sol_rr.kind == RunResultKind.RUNTIME_ERROR,
                time=True,
            )
        )
        text += "\n"
        if judge_rr is not None:
            text += (
                f"{self.judge_name}:\n"
                + tab(
                    self._format_run_result(
                        judge_rr,
                        status=isinstance(self, RunDiffJudge),
                        stderr_force_content=True,
                    )
                )
                + "\n"
            )

        return text

    def verdict_text(self) -> str:
        if self.result is not None:
            if self.result.verdict == Verdict.partial_ok:
                return f"{self.result.verdict.name} {self.result.points:.2f}"
            return self.result.verdict.name
        else:
            return self.state.name

    def verdict_mark(self) -> str:
        if self.state == State.cancelled:
            return "-"
        elif self.result is None:
            return " "
        elif self.result.verdict == Verdict.partial_ok:
            return f"[{self.result.points:.2f}]"
        else:
            return self.result.verdict.mark()


class RunCMSJudge(RunJudge):
    """Judge class with CMS helper functions"""

    def __init__(
        self,
        env: Env,
        judge: TaskPath,
        **kwargs,
    ) -> None:
        super().__init__(env=env, judge_name=judge.name, **kwargs)
        self.judge = judge
        self.points_file = TaskPath.points_file(self._env, self.judge_log_file.name)

    def _load_points(self, result: RunResult) -> float:
        with self._open_file(result.stdout_file) as f:
            points_str = f.read().split("\n")[0]
        try:
            points = float(points_str)
        except ValueError:
            raise self._create_program_failure(
                "Judge didn't write points on stdout:", result
            )

        if not 0 <= points <= 1:
            raise self._create_program_failure(
                "Judge must give between 0 and 1 points:", result
            )

        return points

    def _load_solution_result(self, judge_run_result: RunResult) -> SolutionResult:
        if judge_run_result.returncode == 0:
            points = self._load_points(judge_run_result)
            if points == 1.0:
                verdict = Verdict.ok
            elif points == 0.0:
                verdict = Verdict.wrong_answer
            else:
                verdict = Verdict.partial_ok

            return SolutionResult(
                verdict, points, self._solution_run_res, judge_run_result
            )
        else:
            raise self._create_program_failure(
                f"Judge failed on {self._judging_message()}:", judge_run_result
            )


class RunBatchJudge(RunJudge):
    """Runs batch judge on single input. (Abstract class)"""

    def __init__(
        self,
        env: Env,
        judge_name: str,
        input_: TaskPath,
        output: TaskPath,
        correct_output: TaskPath,
        expected_points: Optional[float],
        **kwargs,
    ) -> None:
        super().__init__(
            env=env,
            name=JUDGE_JOB_NAME.replace(r"(\w+)", output.name, 1),
            judge_name=judge_name,
            input_=input_,
            judge_log_file=TaskPath.log_file(self._env, output.name, judge_name),
            expected_points=expected_points,
            **kwargs,
        )
        self.output = output
        self.correct_output_name = correct_output
        self.correct_output = correct_output

    def _get_solution_run_res(self) -> RunResult:
        if "run_solution" in self.prerequisites_results:
            return self.prerequisites_results["run_solution"]
        else:
            # There is no solution (judging samples)
            # XXX: It didn't technically finish in 0 time.
            return RunResult(RunResultKind.OK, 0, 0.0, 0.0)

    def _judging_message(self) -> str:
        return f"output {self.output:p} for input {self.input:p}"


class RunDiffJudge(RunBatchJudge):
    """Judges solution output and correct output using diff."""

    def __init__(
        self,
        env: Env,
        input_: TaskPath,
        output: TaskPath,
        correct_output: TaskPath,
        expected_points: Optional[float],
    ) -> None:
        super().__init__(
            env=env,
            judge_name="diff",
            input_=input_,
            output=output,
            correct_output=correct_output,
            expected_points=expected_points,
        )

    def _judge(self) -> SolutionResult:
        self._access_file(self.output)
        self._access_file(self.correct_output)
        diff = subprocess.run(
            ["diff", "-Bbq", self.output.path, self.correct_output.path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        # XXX: Okay, it didn't finish in no time, but this is not meant to be used
        rr = RunResult(
            RunResultKind.OK,
            diff.returncode,
            0,
            0,
            status=("Files are the same" if diff.returncode == 0 else "Files differ")
            + f": {self.output.col(self._env)} {self.correct_output.col(self._env)}",
        )
        if diff.returncode == 0:
            return SolutionResult(Verdict.ok, 1.0, self._solution_run_res, rr)
        elif diff.returncode == 1:
            return SolutionResult(Verdict.wrong_answer, 0.0, self._solution_run_res, rr)
        else:
            raise PipelineItemFailure(
                f"Diff failed:\n{tab(diff.stderr.decode('utf-8'))}"
            )


class RunKasiopeaJudge(RunBatchJudge):
    """Judges solution output using judge with Kasiopea interface."""

    def __init__(
        self,
        env: Env,
        judge: TaskPath,
        input_: TaskPath,
        output: TaskPath,
        correct_output: TaskPath,
        subtask: int,
        seed: str,
        expected_points: Optional[float],
        **kwargs,
    ) -> None:
        super().__init__(
            env=env,
            judge_name=judge.name,
            input_=input_,
            output=output,
            correct_output=correct_output,
            expected_points=expected_points,
            **kwargs,
        )
        self.judge = judge
        self.subtask = subtask
        self.seed = seed

    def _judge(self) -> SolutionResult:
        envs = {}
        if self._env.config.judge_needs_in:
            envs["TEST_INPUT"] = self.input.path
            self._access_file(self.input)
        if self._env.config.judge_needs_out:
            envs["TEST_OUTPUT"] = self.correct_output.path
            self._access_file(self.correct_output)

        result = self._run_program(
            ProgramType.judge,
            self.judge,
            args=[str(self.subtask), self.seed],
            stdin=self.output,
            stderr=self.judge_log_file,
            env=envs,
        )
        if result.returncode == 0:
            return SolutionResult(Verdict.ok, 1.0, self._solution_run_res, result)
        elif result.returncode == 1:
            return SolutionResult(
                Verdict.wrong_answer, 0.0, self._solution_run_res, result
            )
        else:
            raise self._create_program_failure(
                f"Judge failed on output {self.output:n}:", result
            )


class RunCMSBatchJudge(RunCMSJudge, RunBatchJudge):
    """Judges solution output using judge with CMS interface."""

    def __init__(
        self,
        env: Env,
        judge: TaskPath,
        input_: TaskPath,
        output: TaskPath,
        correct_output: TaskPath,
        expected_points: Optional[float],
        **kwargs,
    ) -> None:
        super().__init__(
            env=env,
            judge=judge,
            input_=input_,
            output=output,
            correct_output=correct_output,
            expected_points=expected_points,
            **kwargs,
        )

    def _judge(self) -> SolutionResult:
        self._access_file(self.input)
        self._access_file(self.output)
        self._access_file(self.correct_output)
        result = self._run_program(
            ProgramType.judge,
            self.judge,
            args=[
                self.input.path,
                self.correct_output.path,
                self.output.path,
            ],
            stdout=self.points_file,
            stderr=self.judge_log_file,
        )

        sol_result = self._load_solution_result(result)
        return sol_result


def judge_job(
    input_: TaskPath,
    output: TaskPath,
    correct_output: TaskPath,
    subtask: int,
    get_seed: Callable[[], str],
    expected_points: Optional[float],
    env: Env,
) -> Union[RunDiffJudge, RunKasiopeaJudge, RunCMSBatchJudge]:
    """Returns JudgeJob according to contest type."""
    if env.config.out_check == OutCheck.diff:
        return RunDiffJudge(env, input_, output, correct_output, expected_points)

    if env.config.out_judge is None:
        raise RuntimeError(f"Unset judge for out_check={env.config.out_check.name}")
    judge = TaskPath(env.config.out_judge)

    if env.config.judge_type == JudgeType.cms:
        return RunCMSBatchJudge(
            env, judge, input_, output, correct_output, expected_points
        )
    else:
        return RunKasiopeaJudge(
            env,
            judge,
            input_,
            output,
            correct_output,
            subtask,
            get_seed(),
            expected_points,
        )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from enum import Enum
from typing import Optional, Union
import yaml

from pisek.utils.paths import TaskPath


class RunResultKind(Enum):
    OK = 0
    RUNTIME_ERROR = 1
    TIMEOUT = 2


class RunResult:
    """Represents the way the program execution ended. Specially, a program
    that finished successfully, but got Wrong Answer, still gets the OK
    RunResult."""

    def __init__(
        self,
        kind: RunResultKind,
        returncode: int,
        time: float,
        wall_time: float,
        stdout_file: Optional[Union[TaskPath, int]] = None,
        stderr_file: Optional[TaskPath] = None,
        status: str = "",
    ):
        self.kind = kind
        self.returncode = returncode
        self.stdout_file = stdout_file
        self.stderr_file = stderr_file
        self.status = status
        self.time = time
        self.wall_time = wall_time


def run_result_representer(dumper, run_result: RunResult):
    return dumper.represent_sequence(
        "!RunResult",
        [
            run_result.kind.name,
            run_result.returncode,
            run_result.time,
            run_result.wall_time,
            run_result.stdout_file,
            run_result.stderr_file,
            run_result.status,
        ],
    )


def run_result_constructor(loader, value):
    (
        kind,
        returncode,
        time,
        wall_time,
        out_f,
        err_f,
        status,
    ) = loader.construct_sequence(value)
    return RunResult(
        RunResultKind[kind], returncode, time, wall_time, out_f, err_f, status
    )


yaml.add_representer(RunResult, run_result_representer)
yaml.add_constructor("!RunResult", run_result_constructor)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any
import os
from typing import Optional
import yaml

from pisek.utils.text import eprint
from pisek.utils.terminal import colored_env
from pisek.env.env import Env

CACHE_FILENAME = ".pisek_cache"
SAVED_LAST_SIGNATURES = 5


class YAMLObjectWithRequiredAttr(yaml.YAMLObject):
    """Yaml object that has must have all attributes set."""

    @classmethod
    def from_yaml(cls, loader, node):
        return cls(**loader.construct_mapping(node, deep=True))


class CacheEntry(YAMLObjectWithRequiredAttr):
    """Object representing single cached job."""

    yaml_tag = "!Entry"

    def __init__(
        self,
        name: str,
        signature: str,
        result: Any,
        envs: list[str],
        files: list[str],
        prerequisites_results: list[Any],
        output: list[tuple[str, bool]],
    ) -> None:
        self.name = name
        self.signature = signature
        self.result = result
        self.prerequisites_results = list(sorted(prerequisites_results))
        self.envs = list(sorted(envs))
        self.files = list(sorted(files))
        self.output = output

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(name={self.name}, signature={self.signature}, "
            f"result={self.result}, prerequisites_results={self.prerequisites_results}, "
            f"envs={self.envs}, files={self.files}, output={self.output})"
        )

    def yaml_export(self) -> str:
        return yaml.dump([self], allow_unicode=True, sort_keys=False)


class CacheSeal(YAMLObjectWithRequiredAttr):
    """Seal is stored after last entry containing data about run."""

    yaml_tag = "!Seal"

    def __init__(self, success: bool, msg: str = ""):
        self.success = success
        self.msg = msg

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(success={self.success}, msg={self.msg})"

    def yaml_export(self) -> str:
        return yaml.dump([self], allow_unicode=True, sort_keys=False)


class Cache:
    """Object representing all cached jobs."""

    def __init__(self, env: Env) -> None:
        self.broken_seal: Optional[CacheSeal] = None
        self._load(env)

    def add(self, cache_entry: CacheEntry):
        """Add entry to cache."""
        if cache_entry.name not in self.cache:
            self.cache[cache_entry.name] = []
        self.cache[cache_entry.name].append(cache_entry)

        with open(CACHE_FILENAME, "a", encoding="utf-8") as f:
            f.write(cache_entry.yaml_export())

    def __contains__(self, name: str) -> bool:
        return name in self.cache

    def __getitem__(self, name: str) -> list[CacheEntry]:
        return self.cache[name]

    def entry_names(self) -> list[str]:
        return list(self.cache.keys())

    def last_entry(self, name: str) -> CacheEntry:
        return self[name][-1]

    def _load(self, env: Env) -> None:
        """Load cache file."""
        self.cache: dict[str, list[CacheEntry]] = {}
        if not os.path.exists(CACHE_FILENAME):
            return

        with open(CACHE_FILENAME, encoding="utf-8") as f:
            try:
                entries = yaml.full_load(f)
            except (TypeError, ValueError, KeyError):
                eprint(
                    colored_env(
                        "Invalid .pisek_cache file. Starting from scratch...",
                        "yellow",
                        env,
                    )
                )
                entries = {}

            if entries is None:
                return
            for entry in entries:
                if isinstance(entry, CacheSeal):
                    self.broken_seal = entry
                    break
                elif isinstance(entry, CacheEntry):
                    if entry.name not in self.cache:
                        self.cache[entry.name] = []
                    self.cache[entry.name].append(entry)
                else:
                    raise RuntimeError(f"Unknown object in cache {entry}")

        self.export()  # Break the CacheSeal

    def move_to_top(self, entry: CacheEntry):
        """Move given entry to most recent position."""
        if entry in self.cache[entry.name]:
            self.cache[entry.name].remove(entry)
        else:
            raise ValueError(
                f"Cannot move to top entry which is not in Cache:\n{entry}"
            )
        self.add(entry)

    def export(self) -> None:
        """Export cache into a file. Remove unnecessary entries and add seal."""
        with open(CACHE_FILENAME, "w", encoding="utf-8") as f:
            for job, entries in self.cache.items():
                for cache_entry in entries[-SAVED_LAST_SIGNATURES:]:
                    f.write(cache_entry.yaml_export())

    def seal(self, success):
        self.export()
        with open(CACHE_FILENAME, "a", encoding="utf-8") as f:
            f.write(CacheSeal(success).yaml_export())



#!/usr/bin/env python3

from random import Random
import os
import sys

random = Random(275)
directory = sys.argv[1]

for i in range(10):
    nums = set()

    n = random.randint(300000, 400000)
    for _ in range(n):
        nums.add(random.randint(1, 1_000_000))

    nums = list(nums)
    random.shuffle(nums)

    filename = f"{i:02}.in"

    with open(os.path.join(directory, filename), "w") as f:
        for num in nums:
            f.write(f"{num}\n")



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import deque
import os
from typing import Optional
from functools import partial

import pisek.utils.util as util
from pisek.jobs.job_pipeline import JobPipeline
from pisek.utils.pipeline_tools import run_pipeline
from pisek.env.env import Env
from pisek.jobs.cache import Cache

from pisek.jobs.parts.tools import ToolsManager
from pisek.jobs.parts.generator import RunOnlineGeneratorMan
from pisek.jobs.parts.solution import RunPrimarySolutionMan
from pisek.jobs.parts.solution_result import SolutionResult
from pisek.jobs.parts.judge import RunKasiopeaJudgeMan


class KasiopeaInputCase:
    def __init__(self, path: str, subtask: int, seed: int):
        self.path = path

        self.subtask: int = subtask
        self.seed: int = seed

        self.input: Optional[str] = None
        self.correct_output: Optional[str] = None

    def gen_input(self, input_: Optional[str] = None) -> None:
        if input_ is None:
            input_ = util.get_input_name(self.seed, self.subtask)
        self.input = input_

        res = run_pipeline(
            self.path,
            partial(
                ServerGenKasiopea, subtask=self.subtask, seed=self.seed, file=self.input
            ),
        )
        if res != 0:
            raise RuntimeError("Generating input failed.")

    def gen_correct_output(
        self, input_: Optional[str] = None, correct_output: Optional[str] = None
    ) -> None:
        self.input = input_ or self.input
        if self._needs_generating(self.input):
            self.gen_input(self.input)
        assert self.input is not None

        if correct_output is None:
            correct_output = util.get_output_name(self.input, "")
        self.correct_output = correct_output

        res = run_pipeline(
            self.path,
            partial(ServerSolve, input_=self.input, output=self.correct_output),
        )

        if res != 0:
            raise RuntimeError("Generating correct output failed.")

    def judge(
        self,
        output: str,
        input_: Optional[str] = None,
        correct_output: Optional[str] = None,
    ):
        env = Env.load()
        if env is None:
            return 1

        self.input = input_ or self.input
        if env.config.judge_needs_in:
            if self._needs_generating(self.input):
                self.gen_input(self.input)
            assert self.input is not None
        else:
            self.input = "/dev/null"

        self.correct_output = correct_output or self.correct_output
        if env.config.judge_needs_out:
            if self._needs_generating(self.correct_output):
                self.gen_correct_output(self.correct_output)
            assert self.correct_output is not None
        else:
            self.correct_output = "/dev/null"

        pipeline = ServerJudgeKasiopea(
            env.fork(),
            subtask=self.subtask,
            seed=self.seed,
            input_=self.input,
            output=output,
            correct_output=self.correct_output,
        )
        res = pipeline.run_jobs(cache := Cache(env), env)

        if res != 0:
            raise RuntimeError("Judging failed.")

        judging_res = pipeline.judge_result()
        return judging_res.points

    def _needs_generating(self, put: Optional[str]):
        return put is None or not os.path.exists(os.path.join(self.path, "data/", put))


class ServerGenKasiopea(JobPipeline):
    """Generate an input."""

    def __init__(self, env, subtask: int, seed: int, file: str):
        super().__init__()
        if env.config.contest_type == "cms":
            raise NotImplementedError("RunGen for cms is not implemented.")

        self.pipeline = deque(
            [
                tools := ToolsManager(),
                generator := RunOnlineGeneratorMan(subtask, seed, file),
            ]
        )
        generator.add_prerequisite(tools)


class ServerSolve(JobPipeline):
    """Run a primary solution."""

    def __init__(self, env, input_: str, output: Optional[str]):
        super().__init__()

        self.pipeline = deque(
            [
                tools := ToolsManager(),
                solve := RunPrimarySolutionMan(input_, output),
            ]
        )
        solve.add_prerequisite(tools)


class ServerJudgeKasiopea(JobPipeline):
    """Run a primary solution."""

    def __init__(
        self,
        env,
        subtask: int,
        seed: int,
        input_: str,
        output: str,
        correct_output: str,
    ):
        super().__init__()

        self.pipeline = deque(
            [
                tools := ToolsManager(),
                judge := RunKasiopeaJudgeMan(
                    subtask, seed, input_, output, correct_output
                ),
            ]
        )
        judge.add_prerequisite(tools)
        self._judge_man = judge

    def judge_result(self) -> SolutionResult:
        return self._judge_man.judge_result()



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from enum import Enum
from functools import partial, cache
from typing import Callable, Optional
import yaml

from pisek.jobs.parts.run_result import RunResult


class Verdict(Enum):
    ok = 1
    partial_ok = 2
    timeout = 3
    wrong_answer = 4
    error = 5

    def mark(self) -> str:
        return {
            Verdict.ok: "·",
            Verdict.partial_ok: "P",
            Verdict.timeout: "T",
            Verdict.wrong_answer: "W",
            Verdict.error: "!",
        }[self]

    @staticmethod
    @cache
    def pad_length() -> int:
        return max(len(v.name) for v in Verdict)


@dataclass
class SolutionResult:
    """Class representing result of a solution on given input."""

    verdict: Verdict
    points: float
    solution_rr: RunResult
    judge_rr: Optional[RunResult]


def sol_result_representer(dumper, sol_result: SolutionResult):
    return dumper.represent_sequence(
        "!SolutionResult",
        [
            sol_result.verdict.name,
            sol_result.points,
            sol_result.solution_rr,
            sol_result.judge_rr,
        ],
    )


def sol_result_constructor(loader, value) -> SolutionResult:
    verdict, points, sol_rr, judge_rr = loader.construct_sequence(value)
    return SolutionResult(Verdict[verdict], points, sol_rr, judge_rr)


yaml.add_representer(SolutionResult, sol_result_representer)
yaml.add_constructor("!SolutionResult", sol_result_constructor)


def verdict_always(res: Verdict) -> bool:
    return True


def verdict_1point(res: Verdict) -> bool:
    return res == Verdict.ok


def verdict_0points(res: Verdict) -> bool:
    return res in (Verdict.wrong_answer, Verdict.timeout, Verdict.error)


def specific_verdict(res: Verdict, verdict: Verdict) -> bool:
    return res == verdict


# Specifies how expected str should be interpreted
# First function must be true for all
# Second function must be true for any/all according to scoring (min/equal)
SUBTASK_SPEC: dict[str, tuple[Callable[[Verdict], bool], Callable[[Verdict], bool]]] = {
    "1": (verdict_1point, verdict_always),
    "0": (verdict_always, verdict_0points),
    "X": (verdict_always, verdict_always),
    "P": (
        lambda r: not verdict_0points(r),
        partial(specific_verdict, verdict=Verdict.partial_ok),
    ),
    "W": (
        verdict_always,
        partial(specific_verdict, verdict=Verdict.wrong_answer),
    ),
    "!": (verdict_always, partial(specific_verdict, verdict=Verdict.error)),
    "T": (verdict_always, partial(specific_verdict, verdict=Verdict.timeout)),
}



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import os
from typing import Optional
import sys
import signal

from pisek.jobs.task_pipeline import TaskPipeline
from pisek.utils.pipeline_tools import run_pipeline, PATH, locked_folder

from pisek.utils.util import clean_task_dir
from pisek.utils.text import eprint
from pisek.license import license, license_gnu
from pisek.visualize import visualize
from pisek.version import print_version


def sigint_handler(sig, frame):
    eprint("\rStopping...")
    sys.exit(2)


@locked_folder
def test_task(args, **kwargs):
    return test_task_path(PATH, **vars(args), **kwargs)


def test_task_path(path, solutions: Optional[list[str]] = None, **env_args):
    return run_pipeline(path, TaskPipeline, solutions=solutions, **env_args)


def test_solution(args):
    if args.solution is None:
        eprint("Specify a solution name to test.")
        eprint("Example:   pisek [--all_tests] test solution solve_slow_4b")
        return 1

    eprint(f"Testing solution: {args.solution}")
    return test_task(args, solutions=[args.solution])


def test_generator(args):
    eprint("Testing generator")
    return test_task(args, solutions=[])


@locked_folder
def clean_directory(args) -> bool:
    task_dir = PATH
    eprint(f"Cleaning directory: {os.path.abspath(task_dir)}")
    return clean_task_dir(task_dir, args.pisek_dir)


def main(argv):
    parser = argparse.ArgumentParser(
        description=(
            "Tool for developing tasks for programming competitions. "
            "Full documentation is at https://github.com/kasiopea-org/pisek"
        )
    )

    def add_argument_verbosity(parser):
        parser.add_argument(
            "--verbosity",
            "-v",
            action="count",
            default=0,
            help="Be more verbose. (Enter multiple times for even more verbosity.)",
        )

    def add_argument_file_contents(parser):
        parser.add_argument(
            "--file-contents",
            "-C",
            action="store_true",
            help="Show file contents on error",
        )

    def add_argument_timeout(parser):
        parser.add_argument(
            "--timeout",
            "-t",
            type=float,
            help="Override time limit for solutions to TIMEOUT seconds.",
        )

    def add_argument_full(parser):
        parser.add_argument(
            "--full", "-f", action="store_true", help="Don't stop on first failure."
        )

    def add_argument_strict(parser):
        parser.add_argument(
            "--strict",
            action="store_true",
            help="Interpret warnings as failures (for final check).",
        )

    def add_argument_ninputs(parser):
        parser.add_argument(
            "--inputs",
            "-n",
            type=int,
            default=5,
            help="Test on INPUTS difference inputs (only for Kasiopea tasks).",
        )

    def add_argument_all_inputs(parser):
        parser.add_argument(
            "--all-inputs",
            "-a",
            action="store_true",
            help="Test each solution on all inputs.",
        )

    def add_argument_skip_on_timeout(parser):
        parser.add_argument(
            "--skip-on-timeout",
            action="store_true",
            help="Skip all following inputs on first timeout.",
        )

    def add_argument_testing_log(parser):
        parser.add_argument(
            "--testing-log",
            action="store_true",
            help="Write test results to testing_log.json.",
        )

    def add_argument_clean(parser):
        parser.add_argument(
            "--clean",
            "-c",
            action="store_true",
            help="Clean directory beforehand.",
        )

    def add_argument_plain(parser):
        parser.add_argument(
            "--plain",
            action="store_true",
            help="Do not use ANSI escape sequences.",
        )

    def add_argument_no_jumps(parser):
        parser.add_argument(
            "--no-jumps",
            action="store_true",
            help="Do not use ANSI control sequences.",
        )

    def add_argument_no_colors(parser):
        parser.add_argument(
            "--no-colors",
            action="store_true",
            help="Do not use ANSI color sequences.",
        )

    def add_argument_pisek_dir(parser):
        parser.add_argument(
            "--pisek-dir",
            help="Pisek directory where to load configs from. (If not provided uses [git root]/pisek/)",
            type=str,
        )

    def add_argument_filter(parser):
        parser.add_argument(
            "--filter",
            "-f",
            choices=("slowest", "all"),
            default="slowest",
            type=str,
            help="Which inputs to show:\n slowest: Show slowest input only.\n all: Show all inputs.",
        )

    def add_argument_bundle(parser):
        parser.add_argument(
            "--bundle",
            "-b",
            action="store_true",
            help="Don't group inputs by subtask.",
        )

    def add_argument_solutions(parser):
        parser.add_argument(
            "--solutions",
            "-s",
            default=None,
            type=str,
            nargs="*",
            help="Visualize only solutions with a name or source in SOLUTIONS.",
        )

    def add_argument_filename(parser):
        parser.add_argument(
            "--filename",
            default="testing_log.json",
            type=str,
            help="Read testing log from FILENAME.",
        )

    def add_argument_limit(parser):
        parser.add_argument(
            "--limit",
            "-l",
            default=None,
            type=float,
            help="Visualize as if the time limit was LIMIT seconds.",
        )

    def add_argument_segments(parser):
        parser.add_argument(
            "--segments",
            "-S",
            type=int,
            help="Print bars SEGMENTS characters wide.",
        )

    def add_argument_description(parser):
        parser.add_argument(
            "--description",
            "-d",
            type=str,
            help="Created the dataset with a description of DESCRIPTION.",
        )

    def add_argument_dataset(parser):
        parser.add_argument(
            "--dataset",
            "-d",
            type=str,
            required=False,
            help="Use the dataset with the description DESCRIPTION.",
        )

    add_argument_verbosity(parser)
    add_argument_file_contents(parser)
    add_argument_timeout(parser)
    add_argument_full(parser)
    add_argument_strict(parser)
    add_argument_ninputs(parser)
    add_argument_all_inputs(parser)
    add_argument_skip_on_timeout(parser)
    add_argument_clean(parser)
    add_argument_plain(parser)
    add_argument_no_jumps(parser)
    add_argument_no_colors(parser)
    add_argument_testing_log(parser)
    add_argument_pisek_dir(parser)

    subparsers = parser.add_subparsers(help="Run this subcommand.", dest="subcommand")

    parser_version = subparsers.add_parser("version", help="Print current version.")

    parser_test = subparsers.add_parser("test", help="test")
    parser_test.add_argument(
        "target",
        choices=["solution", "generator"],
        help="Test this program type ('solution' or 'generator').",
    )
    parser_test.add_argument(
        "solution", type=str, help="Test the solution with this name.", nargs="?"
    )
    add_argument_timeout(parser_test)
    add_argument_full(parser_test)
    add_argument_strict(parser_test)
    add_argument_ninputs(parser_test)
    add_argument_clean(parser_test)

    _parser_clean = subparsers.add_parser("clean", help="Clean the directory.")

    parser_visualize = subparsers.add_parser(
        "visualize", help="Show solutions statistics and closeness to limit."
    )
    add_argument_filter(parser_visualize)
    add_argument_bundle(parser_visualize)
    add_argument_solutions(parser_visualize)
    add_argument_filename(parser_visualize)
    add_argument_limit(parser_visualize)
    add_argument_segments(parser_visualize)

    parser_license = subparsers.add_parser("license", help="Print license")
    parser_license.add_argument(
        "--print", action="store_true", help="Print entire license."
    )

    parser_cms = subparsers.add_parser("cms", help="Import tasks into CMS.")
    subparsers_cms = parser_cms.add_subparsers(
        help="The subcommand to run.", dest="cms_subcommand"
    )

    parser_cms_create = subparsers_cms.add_parser("create", help="Create a new task.")
    add_argument_description(parser_cms_create)

    parser_cms_update = subparsers_cms.add_parser(
        "update", help="Update the basic properties of an existing task."
    )

    parser_cms_add = subparsers_cms.add_parser(
        "add", help="Add a dataset to an existing task."
    )
    add_argument_description(parser_cms_add)
    parser_cms_add.add_argument(
        "--no-autojudge",
        action="store_true",
        help="Disable background judging for the new dataset.",
    )

    parser_cms_submit = subparsers_cms.add_parser(
        "submit", help="Submit reference solutions for evaluation using CMS."
    )
    parser_cms_submit.add_argument(
        "--username",
        "-u",
        type=str,
        required=True,
        help="Submit as the user with username USERNAME.",
    )

    parser_cms_testing_log = subparsers_cms.add_parser(
        "testing-log",
        help="Generate a testing log for reference solutions submitted to CMS.",
    )
    add_argument_dataset(parser_cms_testing_log)

    parser_cms_check = subparsers_cms.add_parser(
        "check",
        help="Check if reference solutions scored as expected in CMS.",
    )
    add_argument_dataset(parser_cms_check)

    args = parser.parse_args(argv)

    result = None

    if args.clean:
        if not clean_directory(args):
            return 1

    if args.subcommand == "version":
        result = print_version()
    elif args.subcommand == "test":
        if args.target == "solution":
            result = test_solution(args)
        elif args.target == "generator":
            result = test_generator(args)
        else:
            eprint(f"Unknown testing target: {args.target}")
            exit(1)
    elif args.subcommand is None:
        result = test_task(args, solutions=None, target="all")
    elif args.subcommand == "cms":
        args, unknown_args = parser.parse_known_args()

        try:
            import pisek.cms as cms
        except ImportError as err:
            err.add_note("Failed to locate CMS installation")
            raise

        if args.cms_subcommand == "create":
            result = cms.create(args)
        elif args.cms_subcommand == "update":
            result = cms.update(args)
        elif args.cms_subcommand == "add":
            result = cms.add(args)
        elif args.cms_subcommand == "submit":
            result = cms.submit(args)
        elif args.cms_subcommand == "testing-log":
            result = cms.testing_log(args)
        elif args.cms_subcommand == "check":
            result = cms.check(args)
        else:
            raise RuntimeError(f"Unknown CMS command {args.cms_subcommand}")

    elif args.subcommand == "clean":
        result = not clean_directory(args)
    elif args.subcommand == "visualize":
        result = visualize(PATH, **vars(args))
    elif args.subcommand == "license":
        print(license_gnu if args.print else license)
    else:
        raise RuntimeError(f"Unknown subcommand {args.subcommand}")

    return result


def main_wrapped():
    signal.signal(signal.SIGINT, sigint_handler)
    result = main(sys.argv[1:])

    if result:
        exit(1)


if __name__ == "__main__":
    main_wrapped()



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import filecmp
import glob
from math import ceil
import os
import shutil
from typing import Any, Callable, Iterable, Literal

import subprocess
from pisek.env.env import Env
from pisek.config.task_config import ProgramLimits
from pisek.utils.paths import TaskPath
from pisek.utils.terminal import colored_env
from pisek.utils.text import tab
from pisek.config.task_config import SubtaskConfig, ProgramType
from pisek.jobs.jobs import Job
from pisek.jobs.status import StatusJobManager


TOOLS_MAN_CODE = "tools"
GENERATOR_MAN_CODE = "generator"
INPUTS_MAN_CODE = "inputs"
CHECKER_MAN_CODE = "checker"
JUDGE_MAN_CODE = "judge"
SOLUTION_MAN_CODE = "solution_"
DATA_MAN_CODE = "data"


class TaskHelper:
    _env: Env

    def _get_seed(self, input_name: str):
        """Get seed from input name."""
        parts = os.path.splitext(os.path.basename(input_name))[0].split("_")
        if len(parts) == 1:
            return "0"
        else:
            return parts[-1]

    def _get_limits(self, program_type: ProgramType) -> dict[str, Any]:
        """Get execution limits for given program type."""
        limits: ProgramLimits = getattr(self._env.config.limits, program_type.name)
        time_limit = limits.time_limit

        if program_type in (ProgramType.solve, ProgramType.sec_solve):
            if self._env.timeout is not None:
                time_limit = self._env.timeout

        return {
            "time_limit": time_limit,
            "clock_limit": limits.clock_limit(time_limit),
            "mem_limit": limits.mem_limit,
            "process_limit": limits.process_limit,
        }

    def globs_to_files(
        self, globs: Iterable[str], directory: TaskPath
    ) -> list[TaskPath]:
        """Get files in given directory that match any glob."""
        files: list[str] = sum(
            (glob.glob(g, root_dir=directory.path) for g in globs),
            start=[],
        )
        files = list(sorted(set(files)))
        return [TaskPath.from_abspath(directory.path, file) for file in files]

    @staticmethod
    def _short_text(
        text: str,
        style: Literal["h", "t", "ht"] = "h",
        max_lines: int = 10,
        max_chars: int = 100,
    ) -> str:
        """
        Shorten text to max_lines lines and max_chars on line.
        Keep lines from head / tail / both depending on style.
        """
        s_text = []
        for line in text.split("\n"):
            if len(line) > max_chars:
                line = line[: max_chars - 1] + "…"
            s_text.append(line)
        if len(s_text) < max_lines:
            return "\n".join(s_text)

        tail = max(ceil(("t" in style) * max_lines / len(style)) - 1, 0)
        head = max_lines - tail - 1
        return "\n".join(s_text[:head] + ["[…]"] + (s_text[-tail:] if tail else []))

    @staticmethod
    def makedirs(path: TaskPath, exist_ok: bool = True):
        """Make directories"""
        os.makedirs(path.path, exist_ok=exist_ok)

    @staticmethod
    def make_filedirs(path: TaskPath, exist_ok: bool = True):
        """Make directories for given file"""
        os.makedirs(os.path.dirname(path.path), exist_ok=exist_ok)


class TaskJobManager(StatusJobManager, TaskHelper):
    """JobManager class that implements useful methods"""

    def _get_samples(self) -> list[tuple[TaskPath, TaskPath]]:
        """Returns the list [(sample1.in, sample1.out), …]."""
        ins = self._subtask_inputs(self._env.config.subtasks[0])
        outs = (TaskPath.output_static_file(self._env, inp.name) for inp in ins)
        return list(zip(ins, outs))

    def _all_inputs(self) -> list[TaskPath]:
        """Get all input files"""
        return self.prerequisites_results[INPUTS_MAN_CODE]["inputs"]

    def _subtask_inputs(self, subtask: SubtaskConfig) -> list[TaskPath]:
        """Get all inputs of given subtask."""
        return list(filter(lambda p: subtask.in_subtask(p.name), self._all_inputs()))

    def _subtask_new_inputs(self, subtask: SubtaskConfig) -> list[TaskPath]:
        """Get new inputs of given subtask."""
        return list(
            filter(lambda p: subtask.new_in_subtask(p.name), self._all_inputs())
        )


class TaskJob(Job, TaskHelper):
    """Job class that implements useful methods"""

    @staticmethod
    def _file_access(files: int):
        """Adds first i args as accessed files."""

        def dec(f: Callable[..., Any]) -> Callable[..., Any]:
            def g(self, *args, **kwargs):
                for i in range(files):
                    self._access_file(args[i])
                return f(self, *args, **kwargs)

            return g

        return dec

    @_file_access(1)
    def _open_file(self, filename: TaskPath, mode="r", **kwargs):
        if "w" in mode:
            self.make_filedirs(filename)
        return open(filename.path, mode, **kwargs)

    @_file_access(1)
    def _file_exists(self, filename: TaskPath):
        return os.path.isfile(filename.path)

    @_file_access(1)
    def _file_size(self, filename: TaskPath):
        return os.path.getsize(filename.path)

    @_file_access(1)
    def _file_not_empty(self, filename: TaskPath):
        with self._open_file(filename) as f:
            content = f.read()
        return len(content.strip()) > 0

    @_file_access(2)
    def _copy_file(self, filename: TaskPath, dst: TaskPath):
        self.make_filedirs(dst)
        return shutil.copy(filename.path, dst.path)

    @_file_access(2)
    def _link_file(self, filename: TaskPath, dst: TaskPath, overwrite: bool = False):
        self.make_filedirs(dst)
        if overwrite and os.path.exists(dst.path):
            os.remove(dst.path)
        return os.link(filename.path, dst.path)

    @_file_access(2)
    def _files_equal(self, file_a: TaskPath, file_b: TaskPath) -> bool:
        return filecmp.cmp(file_a.path, file_b.path)

    @_file_access(2)
    def _diff_files(self, file_a: TaskPath, file_b: TaskPath) -> str:
        diff = subprocess.run(
            ["diff", file_a.path, file_b.path, "-Bb", "-u2"],
            stdout=subprocess.PIPE,
        )
        return diff.stdout.decode("utf-8")

    def _quote_file(self, file: TaskPath, **kwargs) -> str:
        """Get shortened file contents"""
        with self._open_file(file) as f:
            return self._short_text(f.read().strip(), **kwargs)

    def _quote_file_with_name(
        self, file: TaskPath, force_content: bool = False, **kwargs
    ) -> str:
        """
        Get file name (with its shortened content if env.file_contents or force_content)
        for printing into terminal
        """
        if force_content or self._env.file_contents:
            return f"{file.col(self._env)}\n{colored_env(tab(self._quote_file(file, **kwargs)), 'yellow', self._env)}\n"
        else:
            return f"{file.col(self._env)}\n"



import os
import shutil

import unittest

from pisek.utils.paths import GENERATED_SUBDIR
from util import TestFixtureVariant, overwrite_file, modify_config
from pisek.config.task_config import load_config


class TestSumCMS(TestFixtureVariant):
    def fixture_path(self):
        return "../fixtures/sum_cms/"


class TestMissingGenerator(TestSumCMS):
    def expecting_success(self):
        return False

    def modify_task(self):
        os.remove(os.path.join(self.task_dir, "gen.py"))


class TestGeneratorDoesNotCreateTests(TestSumCMS):
    def expecting_success(self):
        return False

    def modify_task(self):
        overwrite_file(self.task_dir, "gen.py", "gen_dummy.py")


class TestMissingInputFilesForSubtask(TestSumCMS):
    def expecting_success(self):
        return False

    def modify_task(self):
        overwrite_file(self.task_dir, "gen.py", "gen_incomplete.py")


class TestOldInputsDeleted(TestSumCMS):
    """Do we get rid of out-of-date inputs?"""

    def expecting_success(self):
        return False

    def modify_task(self):
        task_config = load_config(self.task_dir)
        self.inputs_dir = os.path.join(
            self.task_dir, task_config.data_subdir, GENERATED_SUBDIR
        )

        # We only care about the generation part, so remove solve.py to stop the tests
        # right after the generator finishes.
        os.remove(os.path.join(self.task_dir, "solve.py"))

        os.makedirs(os.path.join(self.inputs_dir), exist_ok=True)

        with open(os.path.join(self.inputs_dir, "01_outdated.in"), "w") as f:
            # This old input does not conform to the subtask! Get rid of it.
            f.write("-3 -2\n")

    def check_end_state(self):
        self.assertNotIn("01_outdated.in", os.listdir(self.inputs_dir))


class TestScoreCounting(TestSumCMS):
    def expecting_success(self):
        return False

    def modify_task(self):
        overwrite_file(
            self.task_dir, "solve_0b.py", "solve_3b.cpp", new_file_name="solve_0b.cpp"
        )


class TestPartialJudge(TestSumCMS):
    def expecting_success(self):
        return False

    def modify_task(self):
        overwrite_file(self.task_dir, "judge.cpp", "judge_no_partial.cpp")


class TestInvalidJudgeScore(TestSumCMS):
    def expecting_success(self):
        return False

    def modify_task(self):
        overwrite_file(self.task_dir, "judge.cpp", "judge_invalid_score.cpp")


class TestLooseChecker(TestSumCMS):
    """A checker that cannot distinguish between subtasks."""

    def expecting_success(self):
        return False

    def modify_task(self):
        overwrite_file(self.task_dir, "check.py", "check_loose.py")


class TestStrictChecker(TestSumCMS):
    """A checker whose bounds are stricter than what the generator creates."""

    def expecting_success(self):
        return False

    def modify_task(self):
        overwrite_file(self.task_dir, "check.py", "check_strict.py")


class TestDirtySample(TestSumCMS):
    """Sample without newline at the end."""

    def expecting_success(self):
        return False

    def modify_task(self):
        sample = ["3", "1 2", "-8 5", "0 0"]
        with open(os.path.join(self.task_dir, "sample.in"), "w") as f:
            f.write("\n".join(sample))


class TestNoLFInTextInput(TestSumCMS):
    """Input without newline at the end with in_format=text."""

    def expecting_success(self):
        return False

    def modify_task(self):
        def modification_fn(raw_config):
            raw_config["tests"]["in_gen"] = "gen_no_lf"
            raw_config["tests"]["checker"] = ""

        modify_config(self.task_dir, modification_fn)


class TestNoLFInBinaryInput(TestSumCMS):
    """Input without newline at the end with in_format=binary."""

    def expecting_success(self):
        return True

    def modify_task(self):
        def modification_fn(raw_config):
            raw_config["tests"]["in_format"] = "binary"
            raw_config["tests"]["in_gen"] = "gen_no_lf"
            raw_config["tests"]["checker"] = ""

        modify_config(self.task_dir, modification_fn)


class TestNoLFInTextOutput(TestSumCMS):
    """Output without newline at the end with out_format=text."""

    def expecting_success(self):
        return False

    def modify_task(self):
        def modification_fn(raw_config):
            raw_config["solution_solve"]["source"] = "solve_no_lf"
            raw_config["tests"]["checker"] = ""

        modify_config(self.task_dir, modification_fn)


class TestNoLFInBinaryOutput(TestSumCMS):
    """Output without newline at the end with out_format=binary."""

    def expecting_success(self):
        return True

    def modify_task(self):
        def modification_fn(raw_config):
            raw_config["tests"]["out_format"] = "binary"
            raw_config["solution_solve"]["source"] = "solve_no_lf"
            raw_config["tests"]["checker"] = ""

        modify_config(self.task_dir, modification_fn)


class TestGuess(TestFixtureVariant):
    def fixture_path(self):
        return "../fixtures/guess/"


class TestStub(TestFixtureVariant):
    def fixture_path(self):
        return "../fixtures/odd_stub/"


class TestBigInput(TestStub):
    def expecting_success(self):
        return False

    def modify_task(self):
        def modification_fn(raw_config):
            raw_config["limits"]["input_max_size"] = "1"

        modify_config(self.task_dir, modification_fn)


class TestBigOutput(TestStub):
    def expecting_success(self):
        return False

    def modify_task(self):
        def modification_fn(raw_config):
            raw_config["limits"]["output_max_size"] = "1"

        modify_config(self.task_dir, modification_fn)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from configparser import ConfigParser
from copy import copy
import glob
from itertools import product
import os
import re

from pisek.utils.text import eprint, colored
from pisek.config.config_errors import TaskConfigError
from pisek.config.config_types import ProgramType


def rename_key(config: ConfigParser, section: str, key_from: str, key_to: str):
    if key_from in config[section]:
        config[section][key_to] = config[section][key_from]
        del config[section][key_from]


def move_key(config: ConfigParser, key: str, section_from: str, section_to: str):
    if key in config[section_from]:
        if section_to not in config:
            config.add_section("all_solutions")

        config[section_to][key] = config[section_from][key]
        del config[section_from][key]


def maybe_delete_key(config: ConfigParser, section: str, key: str):
    if section in config and key in config[section]:
        del config[section][key]


def update_to_v2(config: ConfigParser, task_path: str) -> None:
    config["task"]["version"] = "v2"

    rename_key(config, "task", "samples_subdir", "static_subdir")
    rename_key(config, "tests", "solution_manager", "stub")

    subtask_points = []
    for section in sorted(config.sections()):
        if re.fullmatch(r"test[0-9]{2}", section):
            if "points" not in config[section]:
                raise TaskConfigError(f"Missing key 'points' in section [{section}]")
            subtask_points.append(int(config[section]["points"]))
    if "test00" not in config.sections():
        subtask_points = [0] + subtask_points

    if "solutions" not in config["task"]:
        raise TaskConfigError("Missing key 'solutions' in section [task]")
    solutions = config["task"]["solutions"].split()
    del config["task"]["solutions"]

    for i, solution in enumerate(solutions):
        if match := re.fullmatch(r"(.*?)_([0-9]{1,3}|X)b", solution):
            points = None if match[2] == "X" else int(match[2])
            if len(
                glob.glob(
                    os.path.join(
                        task_path,
                        config["task"].get("solutions_subdir", ""),
                        f"{solution}.*",
                    )
                )
            ):
                source = solution
            else:
                source = match[1]
        else:
            source = solution
            points = sum(subtask_points)

        if points is None:
            subtasks = "X" * len(subtask_points)
        else:
            subtasks = get_subtask_mask(points, subtask_points)

        solution = f"solution_{solution}"
        config.add_section(solution)
        if i == 0:
            config[solution]["primary"] = "yes"
            subtasks = "1" * len(subtask_points)
        config[solution]["source"] = source
        config[solution]["points"] = "X" if points is None else str(points)
        config[solution]["subtasks"] = subtasks

    subtask_inputs = {}
    in_globs_used = False
    for section in config.sections():
        if not (mat := re.fullmatch(r"test([0-9]{2})", section)):
            continue

        num = int(mat[1])
        if "in_globs" not in config[section]:
            if num == 0:
                subtask_inputs[num] = {"sample*.in"}
            else:
                subtask_inputs[num] = {f"{i:02}*.in" for i in range(1, num + 1)}
        else:
            in_globs_used = True
            subtask_inputs[num] = set(
                map(lambda x: x.replace("_*", "*"), config[section]["in_globs"].split())
            )

    last_subtask = max(subtask_inputs.keys())
    for subtask, inputs in subtask_inputs.items():
        if "0*.in" in inputs or "*.in" in inputs:
            subtask_inputs[subtask] = {
                f"{i:02}*.in" for i in range(1, last_subtask + 1)
            }

    # Now we construct ordering of subtask difficulty but without redundant edges

    # First construct edges
    subtask_includes: dict[int, set[int]] = {i: set([]) for i in subtask_inputs}
    for succ_subtask in subtask_inputs:
        for pred_subtask in subtask_inputs:
            if succ_subtask == pred_subtask:
                continue
            if subtask_inputs[succ_subtask] >= subtask_inputs[pred_subtask]:
                subtask_includes[succ_subtask].add(pred_subtask)

    # Then remove redundant edges
    for succ_subtask in subtask_includes:
        for pred_subtask in subtask_includes:
            if succ_subtask == pred_subtask:
                continue
            if pred_subtask in subtask_includes[succ_subtask]:
                subtask_includes[succ_subtask] -= subtask_includes[pred_subtask]

    if in_globs_used:
        for subtask in subtask_includes:
            subtask_section = config[f"test{subtask:02}"]
            subtask_section["predecessors"] = " ".join(
                map(str, subtask_includes[subtask])
            )

            in_globs = copy(subtask_inputs[subtask])
            for pred in subtask_includes[subtask]:
                in_globs -= subtask_inputs[pred]
            subtask_section["in_globs"] = " ".join(sorted(in_globs))


def get_subtask_mask(points, subtasks):
    all_valid = [0] * len(subtasks)

    valid = 0
    for comb in product([0, 1], repeat=len(subtasks)):
        p = sum([comb[i] * subtasks[i] for i in range(len(subtasks))])
        if p == points:
            valid += 1
            for i in range(len(subtasks)):
                all_valid[i] += comb[i]

    if valid == 0:
        return "X" * len(subtasks)

    sub_mask = ""
    for x in all_valid:
        if x == valid:
            sub_mask += "1"
        elif x == 0:
            sub_mask += "0"
        else:
            sub_mask += "X"
    return sub_mask


def update_to_v3(config: ConfigParser, task_path: str) -> None:
    contest_type = config.get("task", "contest_type", fallback="kasiopea")
    if contest_type not in ["kasiopea", "cms"]:
        raise TaskConfigError(f"Invalid contest_type: '{contest_type}'")
    config["task"]["use"] = f"@{contest_type}"
    # TODO: del config["task"]["contest_type"]

    if "limits" not in config:
        config.add_section("limits")
    for program_type in ProgramType:
        config["limits"][f"{program_type}_clock_mul"] = "0"
        config["limits"][f"{program_type}_clock_min"] = config.get(
            "limits", f"{program_type}_clock_limit", fallback="360"
        )
        maybe_delete_key(config, "limits", f"{program_type}_clock_limit")

    move_key(config, "stub", "tests", "all_solutions")
    move_key(config, "headers", "tests", "all_solutions")

    IGNORED_KEYS = [
        ("task", "tests"),
        ("tests", "in_mode"),
        ("tests", "out_mode"),
        ("tests", "online_validity"),
    ]
    for section, key in IGNORED_KEYS:
        maybe_delete_key(config, section, key)

    for section in config.sections():
        if re.match(r"test\d{2}", section):
            maybe_delete_key(config, section, "file_name")


OUTDATED_VERSIONS = {"v1": ("v2", update_to_v2)}
CURRENT_VERSIONS = {"v2": ("v3", update_to_v3)}

NEWEST_VERSION = "v3"
NEWEST_IS_EXPERIMENTAL = True


def update_config(
    config: ConfigParser, task_path: str, infos: bool = True, no_colors: bool = False
) -> None:
    def inform(msg: str):
        if infos:
            eprint(colored(msg, "yellow", no_colors))

    version = config.get("task", "version", fallback="v1")
    if version == NEWEST_VERSION:
        if NEWEST_IS_EXPERIMENTAL:
            inform(
                f"Config format for version {NEWEST_VERSION} is experimental and can be changed"
            )
        return

    if version in OUTDATED_VERSIONS:
        inform(f"Updating config")

    updaters = OUTDATED_VERSIONS | CURRENT_VERSIONS
    if version not in updaters:
        raise TaskConfigError(f"Unknown version of config: {version}")

    while version in updaters:
        version, updater = updaters[version]
        updater(config, task_path)

    if version != NEWEST_VERSION:
        raise RuntimeError("Config updating failed.")



#!/usr/bin/env python3

import random
import sys
import os

INPUTS = 5

test_dir = sys.argv[1]
os.makedirs(test_dir, exist_ok=True)

random.seed(123)

n = random.sample(range(1, 11), INPUTS)
for i in range(INPUTS):
    test_filename = "{}.in".format(i + 1)
    with open(os.path.join(test_dir, test_filename), "w") as f:
        f.write(str(n[i]) + "\n")



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>

from enum import StrEnum, auto


class TaskType(StrEnum):
    batch = auto()
    communication = auto()


class OutCheck(StrEnum):
    diff = auto()
    judge = auto()


class JudgeType(StrEnum):
    cms = auto()
    opendata_v1 = "opendata-v1"


class DataFormat(StrEnum):
    text = auto()
    binary = auto()


class Scoring(StrEnum):
    equal = auto()
    min = auto()


class ProgramType(StrEnum):
    tool = auto()
    in_gen = auto()
    checker = auto()
    solve = auto()
    sec_solve = auto()
    judge = auto()


class CMSFeedbackLevel(StrEnum):
    full = auto()
    restricted = auto()


class CMSScoreMode(StrEnum):
    max = auto()
    max_subtask = auto()
    max_tokened_last = auto()



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import tempfile
import time
from typing import Any, Optional

from pisek.jobs.jobs import State, Job, PipelineItemFailure
from pisek.env.env import Env
from pisek.utils.paths import TaskPath
from pisek.config.config_types import ProgramType, Scoring
from pisek.utils.text import pad, pad_left, tab, POINTS_DEC_PLACES, format_points
from pisek.utils.terminal import MSG_LEN, colored_env, right_aligned_text
from pisek.jobs.parts.verdicts_eval import evaluate_verdicts
from pisek.jobs.parts.task_job import TaskJobManager
from pisek.jobs.parts.program import RunResult, ProgramsJob
from pisek.jobs.parts.compile import Compile
from pisek.jobs.parts.solution_result import Verdict, SolutionResult
from pisek.jobs.parts.judge import judge_job, RunJudge, RunCMSJudge, RunBatchJudge


class SolutionManager(TaskJobManager):
    """Runs a solution and checks if it works as expected."""

    def __init__(self, solution_label: str) -> None:
        self.solution_label: str = solution_label
        self.solution_points: Optional[float] = None
        self.subtasks: list[SubtaskJobGroup] = []
        self._outputs: list[tuple[TaskPath, RunJudge]] = []
        self._subtasks_results: dict[int, float] = {}
        super().__init__(f"Solution {solution_label} Manager")

    def _get_jobs(self) -> list[Job]:
        self.is_primary: bool = self._env.config.solutions[self.solution_label].primary
        self._solution = TaskPath.solution_path(
            self._env, self._env.config.solutions[self.solution_label].source
        )

        jobs: list[Job] = []

        jobs.append(compile_ := Compile(self._env, self._solution, True))
        self._compile_job = compile_

        self._judges: dict[TaskPath, RunJudge] = {}
        for sub_num, sub in self._env.config.subtasks.items():
            self.subtasks.append(SubtaskJobGroup(self._env, sub_num))
            for inp in self._subtask_inputs(sub):
                if inp not in self._judges:
                    run_sol: RunSolution
                    run_judge: RunJudge
                    if self._env.config.task_type == "batch":
                        run_sol, run_judge = self._create_batch_jobs(sub_num, inp)
                        jobs += [run_sol, run_judge]
                        self._outputs.append((run_judge.output, run_judge))

                    elif self._env.config.task_type == "communication":
                        run_sol = run_judge = self._create_communication_jobs(inp)
                        jobs.append(run_sol)

                    self._judges[inp] = run_judge
                    self.subtasks[-1].new_jobs.append(run_judge)
                    self.subtasks[-1].new_run_jobs.append(run_sol)
                else:
                    self.subtasks[-1].previous_jobs.append(self._judges[inp])

        return jobs

    def _create_batch_jobs(
        self, sub_num: int, inp: TaskPath
    ) -> tuple["RunSolution", RunBatchJudge]:
        """Create RunSolution and RunBatchJudge jobs for batch task type."""
        run_solution = RunBatchSolution(
            self._env,
            self._solution,
            self.is_primary,
            inp,
        )
        run_solution.add_prerequisite(self._compile_job)

        if sub_num == 0:
            c_out = TaskPath.output_static_file(self._env, inp.name)
        else:
            primary_sol = self._env.config.solutions[
                self._env.config.primary_solution
            ].source
            c_out = TaskPath.output_file(self._env, inp.name, primary_sol)

        out = TaskPath.output_file(self._env, inp.name, self._solution.name)
        run_judge = judge_job(
            inp,
            out,
            c_out,
            sub_num,
            lambda: self._get_seed(inp.name),
            None,
            self._env,
        )
        run_judge.add_prerequisite(run_solution, name="run_solution")

        return (run_solution, run_judge)

    def _create_communication_jobs(self, inp: TaskPath) -> "RunCommunication":
        """Create RunCommunication job for communication task type."""
        if self._env.config.out_judge is None:
            raise RuntimeError("Unset judge for communication.")

        judge = TaskPath(self._env.config.out_judge)
        return RunCommunication(self._env, self._solution, self.is_primary, judge, inp)

    def _update(self):
        """Cancel running on inputs that can't change anything."""
        expected = self._env.config.solutions[self.solution_label].subtasks

        for subtask in self.subtasks:
            if subtask.definitive(expected[subtask.num]):
                subtask.cancel()

    def _get_status(self) -> str:
        msg = f"Testing {self.solution_label}"
        if self.state == State.cancelled:
            return self._job_bar(msg)

        points_places = len(str(self._env.config.total_points)) + POINTS_DEC_PLACES + 2
        points = f"{format_points(self.solution_points)}p"

        max_time = max((s.slowest_time for s in self.subtasks), default=0)

        if not self.state.finished() or self._env.verbosity == 0:
            points = pad_left(points, points_places)
            header = f"{pad(msg, MSG_LEN-1)} {points}   "
            subtasks_text = "|".join(sub.status_verbosity0() for sub in self.subtasks)
        else:
            header = (
                right_aligned_text(f"{msg}: {points}", f"slowest {max_time:.2f}s")
                + "\n"
            )
            header = colored_env(header, "cyan", self._env)
            subtasks_text = tab(
                "\n".join(
                    sub.status(self.subtasks, self._env.verbosity)
                    for sub in self.subtasks
                )
            )
            if self._env.verbosity == 1:
                subtasks_text += "\n"

        return header + subtasks_text

    def _evaluate(self) -> None:
        """Evaluates whether solution preformed as expected."""
        self.solution_points = 0
        for sub_job in self.subtasks:
            self.solution_points += sub_job.points
            self._subtasks_results[sub_job.num] = sub_job.normalized_points

        solution_conf = self._env.config.solutions[self.solution_label]
        for sub_job in self.subtasks:
            sub_job.as_expected(solution_conf.subtasks[sub_job.num])

        points = solution_conf.points
        above = solution_conf.points_above
        below = solution_conf.points_below

        if points is not None and self.solution_points != points:
            raise PipelineItemFailure(
                f"Solution {self.solution_label} should have gotten {points} but got {self.solution_points} points."
            )
        elif above is not None and self.solution_points < above:
            raise PipelineItemFailure(
                f"Solution {self.solution_label} should have gotten at least {above} but got {self.solution_points} points."
            )
        elif below is not None and self.solution_points > below:
            raise PipelineItemFailure(
                f"Solution {self.solution_label} should have gotten at most {below} but got {self.solution_points} points."
            )

    def _compute_result(self) -> dict[str, Any]:
        result: dict[str, Any] = {}
        result["outputs"] = {
            Verdict.ok: [],
            Verdict.partial_ok: [],
            Verdict.wrong_answer: [],
            Verdict.timeout: [],
            Verdict.error: [],
        }
        for output, job in self._outputs:
            if job.result is not None:
                result["outputs"][job.result.verdict].append(output)

        result["results"] = {}
        for inp in self._judges:
            result["results"][inp] = self._judges[inp].result

        result["subtasks"] = self._subtasks_results

        return result


class SubtaskJobGroup:
    """Groups jobs of a single subtask."""

    def __init__(self, env: Env, num: int) -> None:
        self._env = env
        self.num = num
        self.subtask = env.config.subtasks[num]
        self.new_run_jobs: list[RunSolution] = []
        self.previous_jobs: list[RunJudge] = []
        self.new_jobs: list[RunJudge] = []

    @property
    def all_jobs(self) -> list[RunJudge]:
        return self.previous_jobs + self.new_jobs

    @property
    def normalized_points(self) -> float:
        results = self._finished_job_results(self.all_jobs)
        points = map(lambda r: r.points, results)
        return min(points, default=1.0)

    @property
    def points(self) -> float:
        return self.normalized_points * self.subtask.points

    @property
    def slowest_time(self) -> float:
        results = self._finished_job_results(self.all_jobs)
        times = map(lambda r: r.solution_rr.time, results)
        return max(times, default=0.0)

    def _job_results(self, jobs: list[RunJudge]) -> list[Optional[SolutionResult]]:
        return list(map(lambda j: j.result, jobs))

    def _finished_jobs(self, jobs: list[RunJudge]) -> list[RunJudge]:
        return list(filter(lambda j: j.result is not None, jobs))

    def _finished_job_results(self, jobs: list[RunJudge]) -> list[SolutionResult]:
        filtered = []
        for res in self._job_results(jobs):
            if res is not None:
                filtered.append(res)
        return filtered

    def _judge_verdicts(self, jobs: list[RunJudge]) -> list[Optional[Verdict]]:
        return list(
            map(lambda r: r.verdict if r is not None else None, self._job_results(jobs))
        )

    def _jobs_points(self) -> list[float]:
        return list(
            map(
                lambda r: r.points,
                self._finished_job_results(self.new_jobs + self.previous_jobs),
            )
        )

    def status(
        self, all_subtasks: list["SubtaskJobGroup"], verbosity: Optional[int] = None
    ) -> str:
        verbosity = self._env.verbosity if verbosity is None else verbosity

        if verbosity <= 0:
            return self.status_verbosity0()
        elif verbosity == 1:
            return self.status_verbosity1()
        elif verbosity >= 2:
            return self.status_verbosity2(all_subtasks)

        raise RuntimeError(f"Unknown verbosity {verbosity}")

    def _verdict_summary(self, jobs: list[RunJudge]) -> str:
        text = ""
        verdicts = self._judge_verdicts(jobs)
        for verdict in Verdict:
            count = verdicts.count(verdict)
            if count > 0:
                text += f"{count}{verdict.mark()}"
        return text

    def _verdict_marks(self, jobs: list[RunJudge]) -> str:
        return "".join(job.verdict_mark() for job in jobs)

    def _predecessor_summary(self) -> str:
        predecessor_summary = self._verdict_summary(self.previous_jobs)
        if predecessor_summary:
            return f"({predecessor_summary}) "
        return ""

    def status_verbosity0(self) -> str:
        return f"{self._predecessor_summary()}{self._verdict_marks(self.new_jobs)}"

    def status_verbosity1(self) -> str:
        max_sub_name_len = max(
            len(subtask.name) for subtask in self._env.config.subtasks.values()
        )
        max_sub_points_len = max(
            len(format_points(sub.points)) for sub in self._env.config.subtasks.values()
        )

        subtask_name = pad(self.subtask.name + ":", max_sub_name_len + 1)
        subtask_points = pad_left(format_points(self.points), max_sub_points_len)

        return right_aligned_text(
            f"{subtask_name} {subtask_points}p  "
            f"{self._predecessor_summary()}{self._verdict_marks(self.new_jobs)}",
            f"slowest {self.slowest_time:.2f}s",
            offset=-2,
        )

    def status_verbosity2(self, all_subtasks: list["SubtaskJobGroup"]):
        def subtask_name(num: int) -> str:
            return self._env.config.subtasks[num].name

        text = ""
        max_inp_name_len = max(len(j.input.name) for j in self.new_jobs)
        subtask_info = (
            right_aligned_text(
                f"{self.subtask.name}: {format_points(self.points)}p",
                f"slowest {self.slowest_time:.2f}s",
                offset=-2,
            )
            + "\n"
        )
        text += colored_env(subtask_info, "magenta", self._env)

        max_pred_name_len = max(
            (len(subtask_name(pred)) for pred in self.subtask.all_predecessors),
            default=0,
        )
        for pred in self.subtask.all_predecessors:
            pred_group = all_subtasks[pred]
            text += right_aligned_text(
                tab(
                    f"Predecessor {pad(subtask_name(pred) + ':', max_pred_name_len + 1)}  "
                    f"{pred_group.status_verbosity0()}"
                ),
                f"slowest {pred_group.slowest_time:.2f}s",
                offset=-2,
            )
            text += "\n"

        if len(self.subtask.all_predecessors) and any(
            map(lambda j: j.result, self.new_jobs)
        ):
            text += "\n"

        for job in self.new_jobs:
            if job.result is not None:
                input_verdict = tab(
                    f"{pad(job.input.name + ':', max_inp_name_len+1)} {job.verdict_text()}"
                )
                text += right_aligned_text(
                    input_verdict, f"{job.result.solution_rr.time:.2f}s", offset=-2
                )
                text += "\n"

        return text

    def definitive(self, expected_str: str) -> bool:
        """Checks whether subtask jobs have resulted in outcome that cannot be changed."""
        if self._env.all_inputs:
            return False

        if self._env.skip_on_timeout and Verdict.timeout in self._judge_verdicts(
            self.new_jobs
        ):
            return True

        if expected_str == "X" and self.normalized_points > 0:
            return False  # Cause X is very very special

        return self._as_expected(expected_str)[1]

    def as_expected(self, expected_str: str) -> None:
        """Checks this subtask resulted as expected. Raises PipelineItemFailure otherwise."""
        ok, _, breaker = self._as_expected(expected_str)
        if not ok:
            msg = f"{self.subtask.name} did not result as expected: '{expected_str}'"
            if breaker is not None:
                msg += f"\n{tab(breaker.message())}"
            raise PipelineItemFailure(msg)

    def _as_expected(self, expected_str: str) -> tuple[bool, bool, Optional[RunJudge]]:
        """
        Returns tuple:
            - whether subtask jobs have resulted as expected
            - whether the result is definitive (cannot be changed)
            - a job that makes the result different than expected (if there is one particular)
        """

        jobs = self.new_jobs + (
            [] if self._env.config.scoring == Scoring.equal else self.previous_jobs
        )

        finished_jobs = self._finished_jobs(jobs)
        verdicts = self._finished_job_results(jobs)

        result, definitive, breaker = evaluate_verdicts(
            self._env.config, list(map(lambda r: r.verdict, verdicts)), expected_str
        )

        breaker_job = None if breaker is None else finished_jobs[breaker]

        return result, definitive, breaker_job

    def cancel(self):
        for job in self.new_run_jobs:
            job.cancel()


class RunPrimarySolutionMan(TaskJobManager):
    def __init__(self, input_: str, output: Optional[str]):
        self._input = input_
        self._output = output
        super().__init__("Running primary solution")

    def _get_jobs(self) -> list[Job]:
        solution = TaskPath.solution_path(
            self._env,
            self._env.config.solutions[self._env.config.primary_solution].source,
        )

        jobs: list[Job] = [
            compile := Compile(self._env, solution, True),
            run_solution := RunBatchSolution(
                self._env,
                solution,
                True,
                TaskPath(self._input),
                TaskPath(self._output) if self._output else None,
            ),
        ]
        run_solution.add_prerequisite(compile)

        return jobs


RUN_JOB_NAME = r"Run (.*) on input (.*)"


class RunSolution(ProgramsJob):
    """Runs solution on given input."""

    def __init__(
        self,
        env: Env,
        name: str,
        solution: TaskPath,
        is_primary: bool,
        **kwargs,
    ) -> None:
        super().__init__(env=env, name=name, **kwargs)
        self.solution = solution
        self.is_primary = is_primary

    def _solution_type(self) -> ProgramType:
        return (ProgramType.solve) if self.is_primary else (ProgramType.sec_solve)


class RunBatchSolution(RunSolution):
    def __init__(
        self,
        env: Env,
        solution: TaskPath,
        is_primary: bool,
        input_: TaskPath,
        output: Optional[TaskPath] = None,
        **kwargs,
    ) -> None:
        name = RUN_JOB_NAME.replace(r"(.*)", solution.name, 1).replace(
            r"(.*)", input_.name, 1
        )
        super().__init__(
            env=env, name=name, solution=solution, is_primary=is_primary, **kwargs
        )
        self.input = input_
        self.output = (
            output
            if output
            else TaskPath.output_file(self._env, self.input.name, solution.name)
        )
        self.log_file = TaskPath.log_file(self._env, input_.name, self.solution.name)

    def _run(self) -> RunResult:
        return self._run_program(
            program_type=self._solution_type(),
            program=self.solution,
            stdin=self.input,
            stdout=self.output,
            stderr=self.log_file,
        )


class RunCommunication(RunCMSJudge, RunSolution):
    def __init__(
        self,
        env: Env,
        solution: TaskPath,
        is_primary: bool,
        judge: TaskPath,
        input_: TaskPath,
        expected_points: Optional[float] = None,
        **kwargs,
    ):
        super().__init__(
            env=env,
            name=RUN_JOB_NAME.replace(r"(.*)", solution.name, 1).replace(
                r"(.*)", input_.name, 1
            ),
            judge=judge,
            input_=input_,
            expected_points=expected_points,
            judge_log_file=TaskPath.log_file(
                self._env, input_.name, f"{solution.name}.{judge.name}"
            ),
            solution=solution,
            is_primary=is_primary,
            **kwargs,
        )
        self.solution = solution
        self.sol_log_file = TaskPath.log_file(
            self._env, self.input.name, self.solution.name
        )

    def _get_solution_run_res(self) -> RunResult:
        with tempfile.TemporaryDirectory() as fifo_dir:
            fifo_from_solution = os.path.join(fifo_dir, "solution-to-manager")
            fifo_to_solution = os.path.join(fifo_dir, "manager-to-solution")

            os.mkfifo(fifo_from_solution)
            os.mkfifo(fifo_to_solution)

            pipes = [
                os.open(fifo_from_solution, os.O_RDWR),
                os.open(fifo_to_solution, os.O_RDWR),
                # Open fifos to prevent blocking on future opens
                fd_from_solution := os.open(fifo_from_solution, os.O_WRONLY),
                fd_to_solution := os.open(fifo_to_solution, os.O_RDONLY),
            ]

            self._load_program(
                ProgramType.judge,
                self.judge,
                stdin=self.input,
                stdout=self.points_file,
                stderr=self.judge_log_file,
                args=[fifo_from_solution, fifo_to_solution],
            )

            self._load_program(
                self._solution_type(),
                self.solution,
                stdin=fd_to_solution,
                stdout=fd_from_solution,
                stderr=self.sol_log_file,
            )

            def close_pipes(_):
                time.sleep(0.05)
                for pipe in pipes:
                    os.close(pipe)

            self._load_callback(close_pipes)

            judge_res, sol_res = self._run_programs()
            self._judge_run_result = judge_res

            return sol_res

    def _judge(self) -> SolutionResult:
        return self._load_solution_result(self._judge_run_result)

    def _judging_message(self) -> str:
        return f"solution {self.solution:p} on input {self.input:p}"



#!/usr/bin/env python3
import argparse


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Nacte ze stdin vstupy ulohy a zkontroluje jejich spravnost"
    )
    parser.add_argument("subtask", type=int, help="subtask (indexovany od 1)")
    args = parser.parse_args()

    # Does not do any actual checking.



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pisek.jobs.jobs import Job
from pisek.jobs.parts.task_job import TaskJobManager, SOLUTION_MAN_CODE


class CompletenessCheck(TaskJobManager):
    """Checks task as a whole."""

    def __init__(self):
        super().__init__("Completeness check")

    def _get_jobs(self) -> list[Job]:
        return []

    def _check_solution_succeeds_only_on(
        self, sol_name: str, subtasks: list[int]
    ) -> bool:
        subtasks_res = self.prerequisites_results[f"{SOLUTION_MAN_CODE}{sol_name}"][
            "subtasks"
        ]
        for num in self._env.config.subtasks:
            if num == 0:
                continue  # Skip samples
            if (subtasks_res[num] == 1.0) != (num in subtasks):
                return False
        return True

    def _evaluate(self) -> None:
        if self._env.config.checks.solution_for_each_subtask:
            for num, subtask in self._env.config.subtasks.items():
                if num == 0:
                    continue  # Samples

                ok = False
                for solution in self._env.config.solutions:
                    if self._check_solution_succeeds_only_on(
                        solution, [num] + subtask.all_predecessors
                    ):
                        ok = True
                        break

                if not ok:
                    self._warn(f"{subtask.name} has no dedicated solution")



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from copy import deepcopy
import hashlib
from enum import Enum, auto
from functools import wraps
import sys
from typing import Optional, AbstractSet, MutableSet, Any
import yaml

import os.path
from pisek.jobs.cache import Cache, CacheEntry
from pisek.env.env import Env
from pisek.utils.paths import TaskPath
from pisek.utils.terminal import colored_env


class State(Enum):
    in_queue = auto()
    running = auto()
    succeeded = auto()
    failed = auto()
    cancelled = auto()

    def finished(self) -> bool:
        return self in (State.succeeded, State.failed)


class PipelineItemFailure(Exception):
    pass


class CaptureInitParams:
    """
    Class that stores __init__ args and kwargs of its descendants
    and forks and locks Env given to it. Only does that to the topmost __init__.
    """

    _initialized = False

    def __init_subclass__(cls):
        real_init = cls.__init__

        @wraps(real_init)
        def wrapped_init(self, env: Env, *args, **kwargs):
            if not self._initialized:
                self._args = args
                self._kwargs = kwargs
                self._initialized = True
                self._env = env.fork()
                self._env.lock()
            real_init(self, self._env, *args, **kwargs)

        cls.__init__ = wrapped_init


class PipelineItem(ABC):
    """Generic PipelineItem with state and dependencies."""

    _env: Env
    run_always: bool = False  # Runs even if prerequisites failed

    def __init__(self, name: str) -> None:
        self.name = name
        self.state = State.in_queue
        self.result: Optional[Any] = None
        self.fail_msg = ""
        self.dirty = False  # Prints something to console?

        self.prerequisites = 0
        self.required_by: list[tuple["PipelineItem", Optional[str]]] = []
        self.prerequisites_results: dict[str, Any] = {}

    def _print(self, msg: str, end: str = "\n", stderr: bool = False) -> None:
        """Prints text to stdout/stderr."""
        self.dirty = True
        (sys.stderr if stderr else sys.stdout).write(msg + end)

    def _warn(self, msg: str) -> None:
        if self._env.strict:
            raise PipelineItemFailure(msg)
        else:
            self._print(colored_env(msg, "yellow", self._env))

    def _fail(self, failure: PipelineItemFailure) -> None:
        """End this job in failure."""
        if self.fail_msg != "":
            raise RuntimeError("PipelineItem cannot fail twice.")
        self.state = State.failed
        self.fail_msg = str(failure)

    def cancel(self) -> None:
        """Cancels job and all that require it."""
        if self.state in (State.succeeded, State.cancelled):
            return  # No need to cancel
        self.state = State.cancelled
        for item, _ in self.required_by:
            if not item.run_always:
                item.cancel()

    def _check_prerequisites(self) -> None:
        """Checks if all prerequisites are finished raises error otherwise."""
        if not self.run_always and self.prerequisites > 0:
            raise RuntimeError(
                f"{self.__class__.__name__} {self.name} prerequisites not finished ({self.prerequisites} remaining)."
            )

    def add_prerequisite(
        self, item: "PipelineItem", name: Optional[str] = None
    ) -> None:
        """Adds given PipelineItem as a prerequisite to this job."""
        self.prerequisites += 1
        item.required_by.append((self, name))

    def finish(self) -> None:
        """Notifies PipelineItems that depend on this job."""
        for item, name in self.required_by:
            if item.run_always or self.state == State.succeeded:
                item.prerequisites -= 1
                if name is not None:
                    item.prerequisites_results[name] = deepcopy(self.result)
            else:
                item.cancel()


class Job(PipelineItem, CaptureInitParams):
    """One simple cacheable task in pipeline."""

    _args: list[Any]
    _kwargs: dict[str, Any]

    def __init__(self, env: Env, name: str) -> None:
        self._env = env
        self._accessed_files: MutableSet[str] = set([])
        self._terminal_output: list[tuple[str, bool]] = []
        self.name = name
        super().__init__(name)

    def _print(self, msg: str, end: str = "\n", stderr: bool = False) -> None:
        """Prints text to stdout/stderr and caches it."""
        self._terminal_output.append((msg + end, stderr))
        return super()._print(msg, end, stderr)

    def _access_file(self, filename: TaskPath) -> None:
        """Add file this job depends on."""
        self._accessed_files.add(filename.path)

    def _signature(
        self, envs: AbstractSet[str], paths: AbstractSet[str], results: dict[str, Any]
    ) -> tuple[Optional[str], Optional[str]]:
        """Compute a signature (i.e. hash) of given envs, files and prerequisites results."""
        sign = hashlib.sha256()
        for i, arg in enumerate(self._args):
            sign.update(f"{i}={arg}\n".encode())
        for key, val in self._kwargs.items():
            sign.update(f"{key}={val}\n".encode())

        for key in sorted(envs):
            try:
                value = self._env.get_compound(key)
            except (AttributeError, TypeError):
                return (None, f"Key nonexistent: {key}")
            sign.update(f"{key}={value}\n".encode())

        expanded_files = []
        for path in sorted(paths):
            if os.path.isfile(path):
                expanded_files.append(path)
            else:
                for dir_, _, dir_files in os.walk(path):
                    for path in dir_files:
                        expanded_files.append(os.path.join(dir_, path))

        for file in sorted(expanded_files):
            if not os.path.exists(file):
                return (None, "File nonexistent")
            with open(file, "rb") as f:
                file_sign = hashlib.file_digest(f, "sha256")
            sign.update(f"{file}={file_sign.hexdigest()}\n".encode())

        for name, result in sorted(results.items()):
            sign.update(f"{name}={yaml.dump(result)}".encode())

        return (sign.hexdigest(), None)

    def _find_entry(self, cache_entries: list[CacheEntry]) -> Optional[CacheEntry]:
        """Finds a corresponding CacheEntry for this Job."""
        for cache_entry in cache_entries:
            sign, err = self._signature(
                set(cache_entry.envs),
                set(cache_entry.files),
                self.prerequisites_results,
            )
            if cache_entry.signature == sign:
                return cache_entry
        return None

    def _export(self, result: Any) -> CacheEntry:
        """Export this job into CacheEntry."""
        sign, err = self._signature(
            set(self._env.get_accessed()),
            self._accessed_files,
            self.prerequisites_results,
        )
        if err == "File nonexistent":
            raise RuntimeError(
                f"Cannot compute signature of job {self.name}. "
                f"Check if something else is changing files in task directory."
            )
        elif sign is None:
            raise RuntimeError(
                f"Computing signature of job {self.name} failed:\n  {err}."
            )
        return CacheEntry(
            self.name,
            sign,
            result,
            self._env.get_accessed(),
            list(self._accessed_files),
            list(self.prerequisites_results),
            self._terminal_output,
        )

    def run_job(self, cache: Cache) -> None:
        """Run this job. If result is already in cache use it instead."""
        if not self._initialized:
            raise RuntimeError(
                "Job must be initialized before running it. (call Job.init)"
            )
        if self.state == State.cancelled:
            return None
        self._check_prerequisites()
        self.state = State.running

        cached = False
        if self.name in cache and (entry := self._find_entry(cache[self.name])):
            cached = True
            cache.move_to_top(entry)
            for msg, stderr in entry.output:
                self._print(msg, end="", stderr=stderr)
            self.result = entry.result
        else:
            try:
                self.result = self._run()
            except PipelineItemFailure as failure:
                self._fail(failure)

        if self.state != State.failed:
            if not cached:
                cache.add(self._export(self.result))
            self.state = State.succeeded

    @abstractmethod
    def _run(self):
        """What this job actually does (without all the management)."""
        pass


class JobManager(PipelineItem):
    """Object that can create jobs and compute depending on their results."""

    def create_jobs(self, env: Env) -> list[Job]:
        """Crates this JobManager's jobs."""
        self.result: Optional[dict[str, Any]]
        self._env = env
        if self.state == State.cancelled:
            self.jobs = []
        else:
            self.state = State.running
            self._check_prerequisites()
            try:
                self.jobs = self._get_jobs()
            except PipelineItemFailure as failure:
                self._fail(failure)
                self.jobs = []
        return self.jobs

    @abstractmethod
    def _get_jobs(self) -> list[Job]:
        """Actually creates this JobManager's jobs (without management)."""
        pass

    def _job_states(self) -> tuple[State, ...]:
        """States of this manager's jobs."""
        return tuple(map(lambda j: j.state, self.jobs))

    def _jobs_with_state(self, state: State) -> list[Job]:
        """Filter this manager's jobs by state."""
        return list(filter(lambda j: j.state == state, self.jobs))

    def _update(self) -> None:
        """Override this function for manager-specific."""
        pass

    def update(self) -> str:
        """Update this manager's state according to its jobs and return status."""
        self._update()
        states = self._job_states()
        if self.state in (State.failed, State.cancelled):
            pass
        elif State.in_queue in states or State.running in states:
            self.state = State.running
        elif State.failed in states:
            self.state = State.failed
        else:
            self.state = State.running

        return self._get_status()

    @abstractmethod
    def _get_status(self) -> str:
        """Return status of job manager to be displayed on stdout."""
        return ""

    def ready(self) -> bool:
        """
        Returns whether manager is ready for evaluation.
        (i.e. All of it's jobs have finished)
        """
        return self.state == State.running and len(
            self._jobs_with_state(State.succeeded)
            + self._jobs_with_state(State.cancelled)
        ) == len(self.jobs)

    def any_failed(self) -> bool:
        """Returns whether this manager or its jobs had any failures so far."""
        return (
            self.state == State.failed or len(self._jobs_with_state(State.failed)) > 0
        )

    def failures(self) -> str:
        """Returns failures of failed jobs or manager itself."""
        failed = self._jobs_with_state(State.failed)
        if len(failed):
            failed_msg = "\n".join([f'"{job.name}": {job.fail_msg}' for job in failed])
            return f"{len(failed)} jobs failed:\n{failed_msg}"
        else:
            return f"{self.name} failed:\n{self.fail_msg}"

    def finalize(self) -> str:
        """Finalizes this JobManager - Does final evaluation and returns final status."""

        if self.state == State.running:
            try:
                self._evaluate()
            except PipelineItemFailure as failure:
                self._fail(failure)
            else:
                self.state = State.succeeded

        self.result = self._compute_result()

        super().finish()
        return self._get_status()

    def _evaluate(self) -> None:
        """Decide whether jobs did run as expected and return result."""
        pass

    def _compute_result(self) -> dict[str, Any]:
        """Creates result to be read by other managers."""
        return {}



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import defaultdict
from configparser import ConfigParser
from dataclasses import dataclass
import editdistance
from importlib.resources import files
import os
import re
from typing import Optional, Any

from pisek.utils.text import tab
from pisek.config.config_errors import TaskConfigError
from pisek.config.update_config import update_config

GLOBAL_DEFAULTS = str(files("pisek").joinpath("config/global-defaults"))
V2_DEFAULTS = {
    task_type: str(files("pisek").joinpath(f"config/{task_type}-defaults"))
    for task_type in ["kasiopea", "cms"]
}
V3_DOCUMENTATION = str(files("pisek").joinpath("config/config-v3-documentation"))

CONFIG_FILENAME = "config"


@dataclass
class ConfigValue:
    value: Any
    config: str
    section: str
    key: Optional[str]
    internal: bool = False

    def location(self) -> str:
        text = f"section [{self.section}]"
        if self.key is not None:
            text += f", key '{self.key}'"

        if self.internal:
            text += " (internal pisek value)"
        else:
            text += f" (in config file {os.path.abspath(self.config)})"

        return text


class ConfigHierarchy:
    """Represents hierarchy of config files where last overrides all previous."""

    def __init__(
        self,
        task_path: str,
        no_colors: bool = False,
        pisek_directory: Optional[str] = None,
    ) -> None:
        self._task_path = task_path
        self._pisek_directory = pisek_directory

        self._config_paths: list[str] = []
        self._configs: list[ConfigParser] = []

        self._load_config(os.path.join(task_path, CONFIG_FILENAME))
        self._load_config(GLOBAL_DEFAULTS, info=False)

        self._used_keys: dict[str, set[str]] = defaultdict(set)

    def _load_config(
        self, path: str, no_colors: bool = False, info: bool = True
    ) -> None:
        self._config_paths.append(path)
        self._configs.append(config := ConfigParser())
        if not config.read(path):
            raise TaskConfigError(f"Missing config {path}. Is this task folder?")

        update_config(config, self._task_path, info, no_colors)
        if defaults := config.get("task", "use", fallback=None):
            self._load_config(self._resolve_defaults_config(defaults), no_colors, False)

    def _resolve_defaults_config(self, name: str):
        def load_from_path(path: str) -> str:
            configs_folder = os.path.join(path, "configs")
            if not os.path.exists(configs_folder):
                raise TaskConfigError(f"'{configs_folder}' does not exist.")
            defaults = os.path.join(configs_folder, name)
            if not os.path.exists(defaults):
                raise TaskConfigError(
                    f"Config '{name}' does not exist in: '{configs_folder}'"
                )
            return defaults

        if name.startswith("@"):
            if (default := V2_DEFAULTS.get(name.removeprefix("@"))) is None:
                raise TaskConfigError(f"Unknown special task_type: '{name}'")
            return default

        if self._pisek_directory is not None:
            return load_from_path(os.path.join(self._task_path, self._pisek_directory))

        if "PISEK_DIRECTORY" in os.environ:
            return load_from_path(
                os.path.join(self._task_path, os.environ["PISEK_DIRECTORY"])
            )

        current_path = os.path.abspath(self._task_path)
        while current_path:
            if os.path.exists(os.path.join(current_path, ".git")):
                return load_from_path(os.path.join(current_path, "pisek"))
            step_up = os.path.abspath(os.path.join(current_path, ".."))
            if step_up == current_path:
                break  # topmost directory
            if os.stat(step_up).st_uid != os.stat(current_path).st_uid:
                break  # other user's directory
            current_path = step_up

        return name

    def get(self, section: str, key: str | None) -> ConfigValue:
        return self.get_from_candidates([(section, key)])

    def get_from_candidates(
        self, candidates: list[tuple[str, str | None]]
    ) -> ConfigValue:
        for section, key in candidates:
            if key is None:
                # Defaultdict, so we are creating the section set
                self._used_keys[section]
            else:
                self._used_keys[section].add(key)

        unset: bool = False
        for config_path, config in zip(self._config_paths, self._configs):
            config_path = os.path.basename(config_path)
            for section, key in candidates:
                if key is None:
                    value = section if section in config else None
                else:
                    value = config.get(section, key, fallback=None)

                if value == "!unset":
                    unset = True
                    break
                elif value is not None:
                    return ConfigValue(value, config_path, section, key)
            if unset:
                break

        def msg(section_key: tuple[str, str | None]) -> str:
            section, key = section_key
            if key is None:
                return f"Section [{section}]"
            else:
                return f"Key '{key}' in section [{section}]"

        candidates_str = " or\n".join(map(msg, candidates))
        raise TaskConfigError(f"Unset value for:\n{tab(candidates_str)}")

    def sections(self) -> list[ConfigValue]:
        sections = {
            section: ConfigValue(section, path, section, None)
            for config, path in reversed(list(zip(self._configs, self._config_paths)))
            for section in config.sections()
        }  # We need to use dictionary here because order matters
        return list(sections.values())

    def check_unused_keys(self) -> None:
        """
        Check whether lowest config contains unused section or keys.

        Raises
        ------
        TaskConfigError
            If unused sections or keys are present.
        """
        IGNORED_KEYS = defaultdict(set, {"task": {"version", "use"}})
        for section in self._configs[0].sections():
            if section not in self._used_keys:
                raise TaskConfigError(f"Unexpected section [{section}] in config.")
            for key in self._configs[0][section].keys():
                if key in IGNORED_KEYS[section]:
                    continue
                if key not in self._used_keys[section]:
                    help_msg = self._help_invalid_key(section, key)
                    raise TaskConfigError(
                        f"Unexpected key '{key}' in section [{section}] of config."
                        + (f" {help_msg}" if help_msg else "")
                    )

    def check_todos(self) -> bool:
        """Check whether lowest config contains TODO in comments."""
        with open(self._config_paths[0]) as config:
            for line in config:
                if "#" in line and "TODO" in line.split("#")[1]:
                    return True

        return False

    def _help_invalid_key(self, section: str, key: str) -> str:
        guess = self._guess_key(section, key)
        if guess is None:
            return ""

        return (
            f"(Did you mean '{guess[1]}'"
            + (f" in section [{guess[0]}]" if section != guess[0] else "")
            + "?)"
        )

    def _guess_key(self, section: str, key: str) -> Optional[tuple[str, str]]:
        v3_docs = ConfigParser()
        v3_docs.read(V3_DOCUMENTATION)

        best_guess = None
        best_score = len(key)
        for section2 in v3_docs.sections():
            for key2 in v3_docs[section2].keys():
                score = editdistance.distance(key, key2) + 5 * (section2 != section)
                if score < best_score:
                    best_guess, best_score = (section2, key2), score

        if best_score == 0 or best_guess is None:
            return None
        return best_guess



# pisek cms - Tool for importing tasks from Pisek into CMS.
#
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Iterator, Any, Optional
from cms.db.task import Task, Dataset, Manager
from cms.db.filecacher import FileCacher
from sqlalchemy.orm import Session
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from os import path, listdir
import re
import datetime

from pisek.cms.testcase import create_testcase, get_testcases
from pisek.env.env import Env
from pisek.config.task_config import TaskConfig
from pisek.config.config_types import OutCheck, TaskType
from pisek.utils.paths import TaskPath


def create_dataset(
    session: Session,
    env: Env,
    task: Task,
    description: Optional[str],
    autojudge: bool = True,
) -> Dataset:
    if description is None:
        description = create_description()

    config = env.config

    score_params = get_group_score_parameters(config)

    task_type: str
    task_params: Any

    if config.task_type == TaskType.batch:
        task_type = "Batch"
        task_params = (
            "grader" if config.stub is not None else "alone",
            ("", ""),
            "comparator" if config.out_check == OutCheck.judge else "diff",
        )
    elif config.task_type == TaskType.communication:
        task_type = "Communication"
        task_params = (1, "stub" if config.stub is not None else "alone", "std_io")
    else:
        raise RuntimeError(f"Cannot upload {config.task_type} task to CMS")

    dataset = Dataset(
        description=description,
        autojudge=autojudge,
        task_type=task_type,
        task_type_parameters=task_params,
        score_type="GroupMin",
        score_type_parameters=score_params,
        time_limit=config.cms.time_limit,
        memory_limit=config.cms.mem_limit * 1024 * 1024,
        task=task,
    )

    session.add(dataset)

    files = FileCacher()

    for testcase in get_testcases(env):
        create_testcase(session, files, dataset, *testcase)

    add_judge(session, files, env, dataset)
    add_stubs(session, files, env, dataset)
    add_headers(session, files, env, dataset)

    return dataset


def get_group_score_parameters(config: TaskConfig) -> list[tuple[int, str]]:
    params = []

    for subtask in config.subtasks.values():
        globs = map(strip_input_extention, subtask.all_globs)
        params.append((subtask.points, globs_to_regex(globs)))

    return params


def strip_input_extention(file: str) -> str:
    if not file.endswith(".in"):
        raise RuntimeError(f"Input file {file} does not have a .in extention")

    return file[:-3]


def glob_char_to_regex(c: str) -> str:
    if c == "?":
        return "."
    elif c == "*":
        return ".*"
    else:
        return re.escape(c)


def globs_to_regex(globs: Iterator[str]) -> str:
    patterns = []

    for glob in globs:
        pattern = "".join(map(glob_char_to_regex, glob))
        patterns.append(f"({pattern})")

    return f"^{'|'.join(patterns)}$"


def add_judge(session: Session, files: FileCacher, env: Env, dataset: Dataset):
    config = env.config

    if config.out_check != OutCheck.judge:
        return

    assert config.out_judge is not None

    if config.task_type == TaskType.batch:
        judge_name = "checker"
    elif config.task_type == TaskType.communication:
        judge_name = "manager"

    judge = files.put_file_from_path(
        TaskPath.executable_path(env, path.basename(config.out_judge)).path,
        f"{judge_name} for {config.name}",
    )
    session.add(Manager(dataset=dataset, filename=judge_name, digest=judge))


MISSING_STUB_ERROR = "Language not supported"
ERROR_STUBS = {
    ".c": f"#error {MISSING_STUB_ERROR}",
    ".cpp": f"#error {MISSING_STUB_ERROR}",
    ".cs": f"#error {MISSING_STUB_ERROR}",
    ".hs": f'{{-# LANGUAGE DataKinds #-}} import GHC.TypeLits; type Error = TypeError (Text "{MISSING_STUB_ERROR}")',
    ".java": MISSING_STUB_ERROR,
    ".pas": f"{{$Message Fatal '{MISSING_STUB_ERROR}'}}",
    ".php": f"<?php throw new Exception('{MISSING_STUB_ERROR}'); ?>",
    ".py": MISSING_STUB_ERROR,
    ".rs": f'compile_error!("{MISSING_STUB_ERROR}");',
}


def add_stubs(session: Session, files: FileCacher, env: Env, dataset: Dataset):
    config = env.config

    if config.stub is None:
        return

    if config.task_type == TaskType.batch:
        stub_basename = "grader"
    elif config.task_type == TaskType.communication:
        stub_basename = "stub"

    directory, target_basename = path.split(config.stub)
    directory = path.normpath(directory)

    exts = set()

    for filename in listdir(directory):
        basename, ext = path.splitext(filename)

        if basename != target_basename:
            continue

        stub = files.put_file_from_path(
            path.join(directory, filename),
            f"{stub_basename}{ext} for {config.name}",
        )
        session.add(
            Manager(dataset=dataset, filename=f"{stub_basename}{ext}", digest=stub)
        )

        exts.add(ext)

    for ext, content in ERROR_STUBS.items():
        if ext in exts:
            continue

        stub = files.put_file_content(
            content.encode(), f"{stub_basename}{ext} for {config.name}"
        )
        session.add(
            Manager(dataset=dataset, filename=f"{stub_basename}{ext}", digest=stub)
        )


def add_headers(session: Session, files: FileCacher, env: Env, dataset: Dataset):
    config = env.config

    for header in config.headers:
        basename = path.basename(header)

        header = files.put_file_from_path(
            header, f"Header {basename} for {config.name}"
        )
        session.add(Manager(dataset=dataset, filename=basename, digest=header))


def get_dataset(session: Session, task: Task, description: Optional[str]) -> Dataset:
    if description is None:
        datasets = session.query(Dataset).filter(Dataset.task == task).all()

        if len(datasets) >= 2:
            raise RuntimeError(
                f"The task has multiple datasets: {', '.join(d.description for d in datasets)}"
            )
        else:
            return datasets[0]

    try:
        return (
            session.query(Dataset)
            .filter(Dataset.task == task)
            .filter(Dataset.description == description)
            .one()
        )
    except NoResultFound as e:
        raise RuntimeError(
            f'The task has no dataset with the description "{description}"'
        ) from e


def create_description() -> str:
    date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    return date



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from collections import deque
from colorama import Cursor, ansi
from math import ceil
import sys
import re
import time

from pisek.env.env import Env
from pisek.utils.terminal import terminal_width
from pisek.jobs.jobs import State, PipelineItem, Job, JobManager
from pisek.jobs.cache import Cache


class JobPipeline(ABC):
    """Runs given Jobs and JobManagers according to their prerequisites."""

    @abstractmethod
    def __init__(self):
        self.failed = False
        self._tmp_lines = 0

    def run_jobs(self, cache: Cache, env: Env) -> bool:
        self.job_managers: deque[JobManager] = deque()
        self.pipeline: deque[PipelineItem] = deque(self.pipeline)
        while len(self.pipeline) or len(self.job_managers):
            p_item = self.pipeline.popleft()
            if isinstance(p_item, JobManager):
                self.job_managers.append(p_item)
                self.pipeline.extendleft(reversed(p_item.create_jobs(env)))
            elif isinstance(p_item, Job):
                p_item.run_job(cache)
                p_item.finish()
            else:
                raise TypeError(
                    f"Objects in {self.__class__.__name__} should be either Job or JobManager."
                )

            if p_item.dirty:
                self._tmp_lines = 0

            # we really need to call status_update to update messages
            # Also no logs into env for just writing to stdout
            self.failed |= not self._status_update(env.fork())
            if self.failed and not env.full:
                break

        cache.seal(not self.failed)  # Remove duplicate cache entries and seal
        return self.failed

    def _status_update(self, env: Env) -> bool:
        """Display current progress. Return true if there were no failures."""
        self._clear_print_tmp()
        while len(self.job_managers):
            job_man = self.job_managers.popleft()
            # We are updating job_man's state with this call!
            ongoing_msg = job_man.update()
            if not env.full and job_man.any_failed():
                self._print(ongoing_msg, env)
                self._print(job_man.failures(), env, end="", file=sys.stderr)
                return False
            if job_man.state == State.failed or job_man.ready():
                self._print_tmp(ongoing_msg, env)
                self._print_active_item(job_man, env)

                job_man.dirty = False
                msg = job_man.finalize()
                if job_man.dirty:
                    self._tmp_lines = 0

                if msg:
                    self._print(msg, env)
                if job_man.state == State.failed:
                    self._print(job_man.failures(), env, end="", file=sys.stderr)
                    return False
            elif job_man.state == State.cancelled:
                self._print(ongoing_msg, env)
            else:
                self._print_tmp(ongoing_msg, env)
                self.job_managers.appendleft(job_man)
                break

        if len(self.pipeline):
            self._print_active_item(self.pipeline[0], env)
        return True

    def _clear_print_tmp(self):
        for _ in range(self._tmp_lines):
            print(f"{Cursor.UP()}{ansi.clear_line()}", end="")
        self._tmp_lines = 0

    def _print_active_item(self, p_item: PipelineItem, env: Env):
        t = time.strftime("%H:%M:%S", time.localtime())
        self._print_tmp(f"Active job: {p_item.name} ({t})", env)

    def _print_tmp(self, msg, env: Env, *args, **kwargs):
        """Prints a text to be rewriten latter."""
        ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
        if not env.no_jumps:
            self._tmp_lines += sum(
                max(ceil(len(re.sub(ansi_escape, "", line)) / terminal_width), 1)
                for line in msg.split("\n")
            )
            print(str(msg), *args, **kwargs)

    def _print(self, msg, env: Env, *args, **kwargs):
        """Prints a text."""
        self._clear_print_tmp()
        print(str(msg), *args, **kwargs)



#!/usr/bin/env python3
import os


def main():
    test_input_file = os.getenv("TEST_INPUT")
    test_output_file = os.getenv("TEST_OUTPUT")

    if test_input_file is None or test_output_file is None:
        exit(2)

    with open(test_input_file, "r") as fin:
        with open(test_output_file, "r") as fcorrect:
            t = int(fin.readline())

            for _ in range(t):
                a, b = [int(x) for x in fin.readline().split()]
                c = int(fcorrect.readline())

                contestant = input()

                if contestant.endswith("\r"):
                    # Looks like we didn't handle Windows line-endings correctly!
                    contestant += "oops!"
                contestant = int(contestant)

                assert a + b == c

                if c != contestant:
                    exit(1)

    exit(0)


main()
